# -*- coding: utf-8 -*-
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2021-2024, Yalin Li <mailto.yalin.li@gmail.com>
#
# This module is under the UIUC open-source license. See
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.

import math
import biosteam as bst
from . import (
    default_insolubles,
    InternalCirculationRx, WWTpump,
    compute_stream_COD, format_str, get_BD_dct, get_split_dct, cost_pump,
)
DesignError = bst.exceptions.DesignError

__all__ = ('AnMBR',)

_ft_to_m = 0.3048 # auom('ft').conversion_factor('m')
_ft2_to_m2 = 0.09290 # auom('ft2').conversion_factor('m2')
_ft3_to_m3 = 0.02832 # auom('ft3').conversion_factor('m3')
_ft3_to_gal = 7.4805 # auom('ft3').conversion_factor('gallon')
_m3_to_gal = 264.1721 # auom('m3').conversion_factor('gal')
_cmh_to_mgd = _m3_to_gal * 24 / 1e6 # cubic meter per hour to million gallon per day
_lb_to_kg = 0.4536 # auom('lb').conversion_factor('kg')


# %%

class AnMBR(bst.Unit):
    """
    Anaerobic membrane bioreactor (AnMBR) for wastewater treatment as in
    Shoener et al. [6]_ Some assumptions adopted from Humbird et al. [2]_

    In addition to the anaerobic treatment, an optional second stage can be added,
    which can be aerobic filter or granular activated carbon (GAC).

    Parameters
    ----------
    ins : 
        * [0] influent
        * [1] recycle (optional)
        * [2] NaOCl
        * [3] citric acid
        * [4] bisulfite
        * [5] air (optional)
    outs : 
        * [0] biogas
        * [1] effluent
        * [2] waste sludge
        * [3] air (optional)
    reactor_type : str
        Can either be "CSTR" for continuous stirred tank reactor
        or "AF" for anaerobic filter.
    membrane_configuration : str
        Can either be "cross-flow" or "submerged".
    membrane_type : str
        Can be "hollow fiber" ("submerged" configuration only),
        "flat sheet" (either "cross-flow" or "submerged" configuration),
        or "multi-tube" ("cross-flow" configuration only).
    membrane_material : str
        Can be any of the plastics ("PES", "PVDF", "PET", "PTFE")
        for any of the membrane types ("hollow fiber", "flat sheet", "multi-tube"),
        or "sintered steel" for "flat sheet",
        or "ceramic" for "multi-tube".
    membrane_unit_cost : float
        Cost of membrane, [$/ft2]
    include_aerobic_filter : bool
        Whether to include an aerobic filtration process in this AnMBR,
        can only be True in "AF" (not "CSTR") reactor.
    add_GAC : bool
        If to add granular activated carbon to enhance biomass retention,
        can only be True for the "submerged" configuration.
    include_degassing_membrane : bool
        If to include a degassing membrane to enhance methane
        (generated through the digestion reaction) recovery.
    Y_biogas : float
        Biogas yield, [kg biogas/kg consumed COD].
    Y_biomass : float
        Biomass yield, [kg biomass/kg consumed COD].
    biodegradability : float or dict
        Biodegradability of chemicals,
        when shown as a float, all biodegradable chemicals are assume to have
        the same degradability.
    split : dict
        Component-wise split to the treated water.
        E.g., {'Water':1, 'WWTsludge':0} indicates all of the water goes to
        the treated water and all of the WWTsludge goes to the wasted sludge.
        Default splits (based on the membrane bioreactor in [2]_) will be used
        if not provided.
    sludge_conc : float
        Concentration of biomass in the waste sludge stream, in g/L. 
        Note that the solids content of the effluent should be smaller than the
        solids content of the waste sludge stream.
    T : float
        Temperature of the reactor.
        Will not control temperature if provided as None.
    include_pump_building_cost : bool
        Whether to include the construction cost of pump building.
    include_excavation_cost : bool
        Whether to include the construction cost of excavation.
    kwargs : dict
        Other keyword arguments (e.g., J_max, SGD).

    """
    _N_ins = 6 # influent, recycle (optional), naocl, citric acid, bisulfite, air (optional)
    _N_outs = 4 # biogas, effluent, waste sludge, air (optional)

    # Equipment-related parameters
    F_BM_pumps = 1.18 * (1+0.007/100) # 0.007 is for  miscellaneous costs
    _F_BM_default = {
        'Membrane': 1 + 0.15, # assume 15% for replacement labor
        'Pumps': F_BM_pumps,
        'Pump building': F_BM_pumps,
        'Pump excavation': F_BM_pumps,
        'Blowers': 2 * 1.11,
        'Blower building': 1.11,
        # Set bare module factor to 1 if not otherwise provided
        'Reactor excavation': 1,
        'Wall concrete': 1,
        'Slab concrete': 1,
        'GAC': 1,
        'Air pipes': 1,
        'Degassing membrane': 1,
        }

    _default_equipment_lifetime = {
        'Membrane': 10,
        'Pumps': 15,
        'Blowers': 15,
        }

    _N_train_min = 2
    _cas_per_tank_spare = 2

    _mod_surface_area = {
        'hollow fiber': 370,
        'flat sheet': 1.45/_ft2_to_m2,
        'multi-tube': 32/_ft2_to_m2
        }

    _mod_per_cas = None
    _mod_per_cas_range = {
        'hollow fiber': (30, 48), # min, max
        'flat sheet': (150, 200),
        'multi-tube': (44, 48)
        }

    _cas_per_tank = None
    _cas_per_tank_range = (16, 22)

    _N_blower = 0

    _W_tank = 21
    _D_tank = 12

    _W_dist = 4.5
    _W_eff = 4.5

    _L_well = 8
    _W_well = 8
    _D_well = 12

    _excav_slope = 1.5
    _constr_access = 3

    # Operation-related parameters
    _HRT = 10
    _J_max = 12
    _TMP_dct = {
        'cross-flow': 2.5,
        'submerged': 2.5,
        }
    _TMP_aerobic = None
    _recir_ratio = 2.25 # from the 0.5-4 uniform range in ref [1]
    _v_cross_flow = 1.2 # from the 0.4-2 uniform range in ref [1]
    _v_GAC = 8
    _SGD = 0.625 # from the 0.05-1.2 uniform range in ref [1]
    _AFF = 3.33
    _sludge_conc = 10.5

    _refresh_rxns = InternalCirculationRx._refresh_rxns

    # Other equipment
    auxiliary_unit_names = ('heat_exchanger',)
    _pumps =  ('perm', 'retent', 'recir', 'sludge', 'naocl', 'citric', 'bisulfite',
               'AF', 'AeF')


    def __init__(self, ID='', ins=None, outs=(), thermo=None, *,
                 reactor_type='CSTR',
                 membrane_configuration='cross-flow',
                 membrane_type='multi-tube',
                 membrane_material='ceramic',
                 membrane_unit_cost=8,
                 include_aerobic_filter=False,
                 add_GAC=False,
                 include_degassing_membrane=True,
                 Y_biogas=0.86,
                 Y_biomass=0.05, # from the 0.02-0.08 uniform range in ref [1]
                 biodegradability={},
                 split={},
                 sludge_conc=10.5,
                 T=35+273.15,
                 include_pump_building_cost=False,
                 include_excavation_cost=False,
                 **kwargs):
        bst.Unit.__init__(self, ID, ins, outs, thermo)
        self.reactor_type = reactor_type
        self.include_aerobic_filter = include_aerobic_filter
        self.membrane_configuration = membrane_configuration
        self.membrane_type = membrane_type
        self.membrane_material = membrane_material
        self.membrane_unit_cost = membrane_unit_cost
        self.add_GAC = add_GAC
        self.include_degassing_membrane = include_degassing_membrane
        self.Y_biogas = Y_biogas
        self.Y_biomass = Y_biomass
        self.biodegradability = \
            biodegradability if biodegradability else get_BD_dct(self.chemicals)
        self.split = split if split else get_split_dct(self.chemicals)
        self.sludge_conc = sludge_conc
        self.T = T
        self.include_pump_building_cost = include_pump_building_cost
        self.include_excavation_cost = include_excavation_cost

        # Initialize the attributes
        ID = self.ID
        self._inf = bst.Stream(f'{ID}_inf')
        self._mixed = bst.Stream(f'{ID}_mixed')
        self._retent = bst.Stream(f'{ID}_retent')
        self._recir = bst.Stream(f'{ID}_recir')        
        hx_in = bst.Stream(f'{ID}_hx_in')
        hx_out = bst.Stream(f'{ID}_hx_out')
        # Add '.' in ID for auxiliary units
        self.heat_exchanger = bst.HXutility(ID=f'.{ID}_hx', ins=hx_in, outs=hx_out)
        self._refresh_rxns()

        for k, v in kwargs.items(): setattr(self, k, v)

        self._check_design()


    def _check_design(self):
        reactor_type = self.reactor_type
        m_config = self.membrane_configuration
        m_type = self.membrane_type
        m_material = self.membrane_material

        if reactor_type == 'CSTR':
            if self.include_aerobic_filter:
                raise DesignError('Aerobic filtration cannot be used in CSTR.')

        if m_config == 'submerged':
            if not m_type in ('hollow fiber', 'flat sheet'):
                raise DesignError('Only "hollow fiber" or "flat sheet" is allowed '
                                  'for "submerged" membrane, not {m_type}.')
        else: # cross-flow
            if not m_type in ('flat sheet', 'multi-tube'):
                raise DesignError('Only "flat sheet" or "multi-tube" is allowed '
                                  'for "cross-flow" membrane, not {m_type}.')
            if self.add_GAC:
                raise DesignError('No GAC should be added '
                                  '(i.e., `add_GAC` can only be False'
                                  'for "cross-flow" membrane.')

        plastics = ('PES', 'PVDF', 'PET', 'PTFE')
        if m_type == 'hollow fiber':
            if not m_material in plastics:
                raise DesignError(f'Only plastic materials {plastics} '
                                  'allowed for "hollow fiber" membrane',
                                  f'not "{m_material}".')
        elif m_type == 'flat sheet':
            if not m_material in (*plastics, 'sintered steel'):
                raise DesignError(f'Only plastic materials {plastics} and "sintered steel"'
                                  'allowed for "flat sheet" membrane',
                                  f'not "{m_material}".')
        else: # multi-tube
            if not m_material in (*plastics, 'ceramic'):
                raise DesignError(f'Only plastic materials {plastics} and "ceramic"'
                                  'allowed for "multi-tube" membrane',
                                  f'not "{m_material}".')


    @staticmethod
    def _degassing(original_stream, receiving_stream):
        InternalCirculationRx._degassing(original_stream, receiving_stream)


    @staticmethod
    def compute_COD(stream):
        r"""
        Compute the chemical oxygen demand (COD) of a given stream in kg-O2/m3
        by summing the COD of each chemical in the stream using:

        .. math::
            COD [\frac{kg}{m^3}] = mol_{chemical} [\frac{kmol}{m^3}] * \frac{g O_2}{mol chemical}
        """
        return compute_stream_COD(stream)


    # =========================================================================
    # _run
    # =========================================================================
    def _run(self):
        raw, recycled, naocl, citric, bisulfite, air_in = self.ins
        biogas, perm, sludge, air_out = self.outs
        degassing = self._degassing

        # Initialize the streams
        biogas.phase = 'g'
        biogas.empty()

        mixed = self._mixed
        mixed.mix_from((raw, recycled))
        self._inf.copy_like(mixed) # this stream will be preserved (i.e., no reaction)

        # Chemicals for cleaning, assume all chemicals will be used up
        # 2.2 L/yr/cmd of 12.5 wt% solution (15% vol)
        naocl.empty()
        naocl.imass['NaOCl', 'Water'] = [0.125, 1-0.125]
        naocl.F_vol = (2.2/1e3/365/24) * (mixed.F_vol*24) # m3/hr solution

        # 0.6 L/yr/cmd of 100 wt% solution, 13.8 lb/kg
        citric.empty()
        citric.ivol['CitricAcid'] = (0.6/1e3/365/24) * (mixed.F_vol*24) # m3/hr pure

        # 0.35 L/yr/cmd of 38% solution, 3.5 lb/gal
        bisulfite.empty()
        bisulfite.imass['Bisulfite', 'Water'] = [0.38, 1-0.38]
        bisulfite.F_vol = (0.35/1e3/365/24) * (mixed.F_vol*24) # m3/hr solution

        # For pump design
        self._compute_mod_case_tank_N()
        Q_R_mgd, Q_IR_mgd = self._compute_liq_flows()
        retent, recir = self._retent, self._recir
        retent.copy_like(mixed)
        recir.copy_like(mixed)
        retent.F_mass *= Q_R_mgd / self.Q_mgd
        recir.F_mass *= Q_IR_mgd / self.Q_mgd

        # Effluents
        self.growth_rxns(mixed.mol)
        self.biogas_rxns(mixed.mol)
        mixed.split_to(perm, sludge, self._isplit.data)

        sludge_conc = self._sludge_conc
        insolubles = tuple(i.ID for i in self.chemicals if i.ID in default_insolubles)
        m_insolubles = sludge.imass[insolubles].sum()
        if m_insolubles/sludge.F_vol <= sludge_conc:
            diff = sludge.ivol['Water'] - m_insolubles/sludge_conc
            sludge.ivol['Water'] = m_insolubles/sludge_conc
            perm.ivol['Water'] += diff
        if perm.imass['Water'] < 0:
            raise ValueError('Not enough moisture in the influent for waste sludge '
                             f'with {sludge_conc} g/L sludge concentration.')

        degassing(perm, biogas)
        degassing(sludge, biogas)

        # Gas for sparging, no sparging needed if submerged or using GAC
        air_out.link_with(air_in)
        air_in.T = 17 + 273.15
        self._design_blower()
        
        if self.T is not None: perm.T = sludge.T = biogas.T = air_out.T = self.T


    # Called by _run
    def _compute_mod_case_tank_N(self):
        N_mod_min, N_mod_max = self.mod_per_cas_range[self.membrane_type]
        N_cas_min, N_cas_max = self.cas_per_tank_range

        mod_per_cas, cas_per_tank = N_mod_min, N_cas_min

        J, J_max, N_train = self.J, self.J_max, self._N_train_min
        while J > J_max:
            mod_per_cas += 1
            if mod_per_cas == N_mod_max + 1:
                if cas_per_tank == N_cas_max + 1:
                    N_train += 1
                    mod_per_cas, cas_per_tank = N_mod_min, N_cas_min
                else:
                    cas_per_tank += 1
                    mod_per_cas = N_mod_min

        self._N_train, self._mod_per_cas, self._cas_per_tank = \
            N_train, mod_per_cas, cas_per_tank


    # Called by _run
    def _compute_liq_flows(self):
        m_type = self.membrane_type

        if m_type == 'multi-tube':
            # Cross-flow flow rate per module,
            # based on manufacture specifications for compact 33, [m3/hr]
            Q_cross_flow = 53.5 * self.v_cross_flow
            Q_R_cmh = self.N_mod_tot * Q_cross_flow # total retentate flow rate, [m3/hr]
            Q_R_mgd = Q_R_cmh * _cmh_to_mgd # [mgd]

        Q_mgd, recir_ratio = self.Q_mgd, self.recir_ratio
        if Q_mgd*recir_ratio >= Q_R_mgd:
            Q_IR_mgd = Q_mgd*recir_ratio - Q_R_mgd
        else:
            Q_IR_mgd = 0
            # # Gives really large recir_ration,
            # # probably shouldn't back-calculate this way
            # self._recir_ratio = Q_R_mgd / Q_mgd

        if self.add_GAC:
            Q_upflow_req = (self.v_GAC/_ft_to_m) * \
                self.L_membrane_tank*self.W_tank*self.N_train * 24 * _ft3_to_gal / 1e6
            Q_IR_add_mgd = max(0, (Q_mgd+Q_IR_mgd)-Q_upflow_req)
            Q_IR_mgd += Q_IR_add_mgd
            self._recir_ratio = Q_IR_mgd / Q_mgd

        return Q_R_mgd, Q_IR_mgd


    # Called by _run
    def _design_blower(self):
        air = self.ins[-1]

        if (not self.add_GAC) and (self.membrane_configuration=='submerged'):
            gas = self.SGD * self.mod_surface_area*_ft2_to_m2 # [m3/h]
            gas /= (_ft3_to_m3 * 60) # [ft3/min]
            gas_train = gas * self.N_train*self.cas_per_tank*self.mod_per_cas

            TCFM = math.ceil(gas_train) # total cubic ft per min
            N = 1
            if TCFM <= 30000:
                CFMB = TCFM / N # cubic ft per min per blower
                while CFMB > 7500:
                    N += 1
                    CFMB = TCFM / N
            elif 30000 < TCFM <= 72000:
                CFMB = TCFM / N
                while CFMB > 18000:
                    N += 1
                    CFMB = TCFM / N
            else:
                CFMB = TCFM / N
                while CFMB > 100000:
                    N += 1
                    CFMB = TCFM / N

            gas_m3_hr = TCFM * _ft3_to_m3 * 60 # ft3/min to m3/hr

            air.ivol['N2'] = 0.79
            air.ivol['O2'] = 0.21
            air.F_vol = gas_m3_hr
        else: # no sparging/blower needed
            TCFM = CFMB = 0.
            N = -1 # to account for the spare
            air.empty()

        D = self.design_results
        D['Total air flow [CFM]'] = TCFM
        D['Blower capacity [CFM]'] = CFMB
        D['Blowers'] = self._N_blower = N + 1 # add a spare


    # =========================================================================
    # _design
    # =========================================================================
    def _design(self):
        D = self.design_results     
        D['Treatment train'] = self.N_train
        D['Cassette per train'] = self.cas_per_tank
        D['Module per cassette'] = self.mod_per_cas
        D['Total membrane modules'] = self.N_mod_tot

        # Step A: Reactor and membrane tanks
        # Call the corresponding design function
        # (_design_CSTR or _design_AF)
        func = getattr(self, f'_design_{self.reactor_type}')

        wall, slab, excavation = func()
        D['Wall concrete [ft3]'] = wall
        D['Slab concrete [ft3]'] = slab
        D['Excavation [ft3]'] = excavation

        # Optional addition of packing media (used in filters)
        ldpe, hdpe = 0., 0.
        for i in (self.AF, self.AeF):
            if i is None:
                continue
            ldpe += i.design_results['Packing LDPE [m3]']
            hdpe += i.design_results['Packing HDPE [m3]']

        # Optional addition of GAC
        D['GAC [kg]'] = self._design_GAC()

        # Step B: Membrane
        # Call the corresponding design function
        # (_design_hollow_fiber, _design_flat_sheet, or _design_multi_tube)
        m_type = format_str(self.membrane_type)
        func = getattr(self, f'_design_{m_type}')
        D['Membrane [m3]'] = func()

        # Step C: Pumps
        pipe, pumps, hdpe = self._design_pump()
        D['Pipe stainless steel [kg]'] = pipe
        D['Pump stainless steel [kg]'] = pumps
        D['Pump chemical storage HDPE [m3]'] = hdpe

        # Step D: Degassing membrane
        D['Degassing membrane'] = self.N_degasser

        # Total volume
        D['Total volume [ft3]'] = self.V_tot


    ### Step A functions ###
    # Called by _design
    def _design_CSTR(self):
        N_train = self.N_train
        W_dist, L_CSTR , W_eff, L_membrane_tank = \
            self.W_dist, self.L_CSTR, self.W_eff, self.L_membrane_tank
        W_PB, W_BB, D_tank = self.W_PB, self.W_BB, self.D_tank
        t_wall, t_slab = self.t_wall, self.t_slab
        SL, CA = self.excav_slope, self.constr_access

        ### Concrete calculation ###
        W_N_trains = (self.W_tank+2*t_wall)*N_train - t_wall*(N_train-1)

        D = D_tank + 2 # add 2 ft of freeboard
        t = t_wall + t_slab

        get_VWC = lambda L1, N: N * t_wall * L1 * D
        get_VSC = lambda L2: t * L2 * W_N_trains

        # Concrete for distribution channel, [ft3]
        VWC_dist = get_VWC(L1=(W_N_trains+W_dist), N=2)
        VSC_dist = get_VSC(L2=(W_dist+2*t_wall))

        # Concrete for CSTR tanks, [ft3]
        VWC_CSTR = get_VWC(L1=L_CSTR, N=(N_train+1))
        VSC_CSTR = get_VSC(L2=L_CSTR)

        # Concrete for effluent channel, [ft3]
        VWC_eff = get_VWC(L1=(W_N_trains+W_eff), N=2)
        VSC_eff = get_VSC(L2=(W_eff+2*t_wall))

        # Concrete for the pump/blower building, [ft3]
        VWC_PBB = get_VWC(L1=(W_N_trains+W_PB+W_BB), N=2)
        VSC_PBB = get_VSC(L2=(W_PB+t_wall+W_BB))

        if self.membrane_configuration == 'submerged':
            VWC_membrane_tank, VSC_membrane_tank, VWC_well, VSC_well = \
                self._design_membrane_tank(self, D, N_train, W_N_trains,
                                           self.L_membrane_tank)
        else:
            VWC_membrane_tank, VSC_membrane_tank, VWC_well, VSC_well = 0., 0., 0., 0.

        # Total volume of wall concrete, [ft3]
        VWC = VWC_dist + VWC_CSTR + VWC_eff + VWC_PBB + VWC_membrane_tank + VWC_well

        # Total volume of slab concrete [ft3]
        VSC = VSC_dist + VSC_CSTR + VSC_eff + VSC_PBB + VSC_membrane_tank + VSC_well

        ### Excavation calculation ###
        get_VEX = lambda L_bttom, W_bottom, diff: \
            0.5 * D_tank * (L_bottom*W_bottom+(L_bottom+diff)*(W_bottom+diff)) # bottom+top

        L_bottom = W_dist + L_CSTR + W_eff + L_membrane_tank + 2*CA
        W_bottom = W_N_trains + 2*CA
        diff = D_tank * SL

        # Excavation volume for membrane tanks, [ft3]
        VEX_membrane_tank = get_VEX(L_bottom, W_bottom, diff)

        # Excavation volume for pump/blower building, [ft3]
        VEX_PBB = get_VEX((W_PB+W_BB+2*CA), W_bottom, diff)

        VEX = VEX_membrane_tank + VEX_PBB

        return VWC, VSC, VEX


    # Called by _design
    def _design_AF(self):
        raise RuntimeError('Design not implemented.')


    # Called by _design_CSTR/_design_AF
    def _design_membrane_tank(self, D, N_train, W_N_trains, L_membrane_tank,
                              t_wall, t_slab):
        L_well, W_well, D_well = self.L_well, self.W_well, self.D_well

        # Concrete for membrane tanks, [ft3]
        t = t_wall + t_slab
        VWC_membrane_tank = (N_train+1) * t_wall * L_membrane_tank * D
        VSC_membrane_tank = t * L_membrane_tank * W_N_trains

        # Concrete for wet well (mixed liquor storage), [ft3]
        L = L_well + 2*t_wall
        W = W_well + 2*t_wall

        VWC_well = 2 * t_wall * (L_well+W) * D_well
        VSC_well = (t_slab+t_wall) * L * W

        return VWC_membrane_tank, VSC_membrane_tank, VWC_well, VSC_well


    # Called by _design
    def _design_GAC(self):
        if not self.add_GAC: return 0
        raise RuntimeError('Design not implemented.')


    ### Step B functions ###
    # Called by _design
    def _design_hollow_fiber(self):
        raise RuntimeError('Design not implemented.')


    # Called by _design
    def _design_flat_sheet(self):
        raise RuntimeError('Design not implemented.')


    # Called by _design
    def _design_multi_tube(self):
        # # 0.01478 is volume of material for each membrane tube, [m3]
        # #      L_tube               OD        ID
        # V_tube = 3 * math.pi/4 * ((6e-3)**2-(5.2e-3)**2)
        # V_SU = 700 * V_tube # V for each small unit [m3]
        # M_SU = 1.78*1e3 * V_SU # mass = density*volume, [kg/m3], not used
        return self.N_mod_tot*0.01478


    ### Step C function ###
    # Called by _design
    def _design_pump(self):
        rx_type, m_config, pumps = \
            self.reactor_type, self.membrane_configuration, self._pumps

        ins_dct = {
            'perm': self.outs[1].proxy(),
            'retent': self._retent,
            'recir': self._recir,
            'sludge': self.outs[2].proxy(),
            'naocl': self.ins[2].proxy(),
            'citric': self.ins[3].proxy(),
            'bisulfite': self.ins[4].proxy(),
            }

        type_dct = {
            'perm': f'permeate_{m_config}',
            'retent': f'retentate_{rx_type}',
            'recir': f'recirculation_{rx_type}',
            'sludge': 'sludge',
            'naocl': 'chemical',
            'citric': 'chemical',
            'bisulfite': 'chemical',
            }

        inputs_dct = {
            'perm': (self.cas_per_tank, self.D_tank, self.TMP_anaerobic,
                     self.include_aerobic_filter),
            'retent': (self.cas_per_tank,),
            'recir': (self.L_CSTR,),
            'sludge': (),
            'naocl': (),
            'citric': (),
            'bisulfite': (),
            }

        WWTpump._batch_adding_pump(self, pumps[:-2], ins_dct, type_dct, inputs_dct)

        self.AF_pump = self.AF.lift_pump if self.AF else None
        self.AeF_pump = self.AeF.lift_pump if self.AeF else None

        pipe_ss, pump_ss, hdpe = 0., 0., 0.
        for i in pumps:
            p = getattr(self, f'{i}_pump')
            if p == None: continue

            p.simulate()
            pipe_ss += p.design_results['Pipe stainless steel [kg]']
            pump_ss += p.design_results['Pump stainless steel [kg]']
            hdpe += p.design_results['Chemical storage HDPE [m3]']

        return pipe_ss, pump_ss, hdpe


    # =========================================================================
    # _cost
    # =========================================================================
    def _cost(self):
        D, C = self.design_results, self.baseline_purchase_costs

        # Concrete and excavation
        VEX, VWC, VSC = \
            D['Excavation [ft3]'], D['Wall concrete [ft3]'], D['Slab concrete [ft3]']
        # 27 is to convert the VEX from ft3 to yard3
        C['Reactor excavation'] = VEX/27*8 if self.include_excavation_cost else 0.
        C['Wall concrete'] = VWC / 27 * 650
        C['Slab concrete'] = VSC / 27 * 350

        # Membrane
        C['Membrane'] = self.membrane_unit_cost * D['Membrane [m3]'] / _ft2_to_m2

        # GAC
        # $13.78/kg
        C['GAC'] = 13.78 * D['GAC [kg]']

        # Packing material
        ldpe, hdpe = 0., 0.
        for i in (self.AF, self.AeF):
            if i is None:
                continue
            ldpe += i.baseline_purchase_costs['Packing LDPE [m3]']
            hdpe += i.baseline_purchase_costs['Packing HDPE [m3]']

        # Heat loss, assume air is 17°C, ground is 10°C
        T = self.T
        if T is None: loss = 0.
        else:
            N_train, L_CSTR, W_tank, D_tank = \
                self.N_train, self.L_CSTR, self.W_tank, self.D_tank
            A_W = 2 * (L_CSTR+W_tank) * D_tank
            A_F = L_CSTR * W_tank
            A_W *= N_train * _ft2_to_m2
            A_F *= N_train * _ft2_to_m2

            loss = 0.7 * (T-(17+273.15)) * A_W # 0.7 W/m2/°C for wall
            loss += 1.7 * (T-(10+273.15)) * A_F # 1.7 W/m2/°C for floor
            loss += 0.95 * (T-(17+273.15)) * A_F # 0.95 W/m2/°C for floating cover
            loss *= 3.6 # W (J/s) to kJ/hr

        # Stream heating
        hx = self.heat_exchanger
        inf = self._inf
        hx_ins0, hx_outs0 = hx.ins[0], hx.outs[0]
        hx_ins0.copy_flow(inf)
        hx_outs0.copy_flow(inf)
        hx_ins0.T = inf.T
        hx_outs0.T = T
        hx.H = hx_outs0.H + loss # stream heating and heat loss
        hx.simulate_as_auxiliary_exchanger(ins=hx.ins, outs=hx.outs)

        # Pump
        # Note that maintenance and operating costs are included as a lumped
        # number in the biorefinery thus not included here
        pumps, building = cost_pump(self)
        C['Pumps'] = pumps
        C['Pump building'] = building if self.include_pump_building_cost else 0.
        C['Pump excavation'] = VEX/27*0.3 if self.include_excavation_cost else 0.

        # Blower and air pipe
        TCFM, CFMB = D['Total air flow [CFM]'], D['Blower capacity [CFM]']
        C['Air pipes'], C['Blowers'], C['Blower building'] = self._cost_blower(TCFM, CFMB)

        # Degassing membrane
        C['Degassing membrane'] = 10000 * D['Degassing membrane']

        # Pumping
        pumping = 0.
        for ID in self._pumps:
            p = getattr(self, f'{ID}_pump')
            if p is None: continue
            pumping += p.power_utility.rate

        # Gas
        sparging = 0. # submerge design not implemented
        degassing = 3 * self.N_degasser # assume each uses 3 kW

        self.power_utility.rate = sparging + degassing + pumping


    # Called by _cost
    def _cost_blower(self, TCFM, CFMB):
        AFF = self.AFF

        # Air pipes
        # Note that the original codes use CFMD instead of TCFM for air pipes,
        # but based on the coding they are equivalent
        if TCFM <= 1000:
            air_pipes = 617.2 * AFF * (TCFM**0.2553)
        elif 1000 < TCFM <= 10000:
            air_pipes = 1.43 * AFF * (TCFM**1.1337)
        else:
            air_pipes = 28.59 * AFF * (TCFM**0.8085)

        # Blowers
        if TCFM <= 30000:
            ratio = 0.7 * (CFMB**0.6169)
            blowers = 58000*ratio / 100
        elif 30000 < TCFM <= 72000:
            ratio = 0.377 * (CFMB**0.5928)
            blowers = 218000*ratio / 100
        else:
            ratio = 0.964 * (CFMB**0.4286)
            blowers  = 480000*ratio / 100

        # Blower building
        area = 128 * (TCFM**0.256) # building area, [ft2]
        building = area * 90 # 90 is the unit price, [USD/ft2]

        return air_pipes, blowers, building


    ### Reactor configuration ###
    @property
    def reactor_type(self):
        """
        [str] Can either be "CSTR" for continuous stirred tank reactor
        or "AF" for anaerobic filter.
        """
        return self._reactor_type
    @reactor_type.setter
    def reactor_type(self, i):
        if not i.upper() in ('CSTR', 'AF'):
            raise ValueError('`reactor_type` can only be "CSTR", or "AF", '
                             f'not "{i}".')
        self._reactor_type = i.upper()

    @property
    def membrane_configuration(self):
        """[str] Can either be "cross-flow" or "submerged"."""
        return self._membrane_configuration
    @membrane_configuration.setter
    def membrane_configuration(self, i):
        i = 'cross-flow' if i.lower() in ('cross flow', 'crossflow') else i
        if not i.lower() in ('cross-flow', 'submerged'):
            raise ValueError('`membrane_configuration` can only be "cross-flow", '
                             f'or "submerged", not "{i}".')
        self._membrane_configuration = i.lower()

    @property
    def membrane_type(self):
        """
        [str] Can be "hollow fiber" ("submerged" configuration only),
        "flat sheet" (either "cross-flow" or "submerged" configuration),
        or "multi-tube" ("cross-flow" configuration only).
        """
        return self._membrane_type
    @membrane_type.setter
    def membrane_type(self, i):
        i = 'multi-tube' if i.lower() in ('multi tube', 'multitube') else i
        if not i.lower() in ('hollow fiber', 'flat sheet', 'multi-tube'):
            raise ValueError('`membrane_type` can only be "hollow fiber", '
                             f'"flat sheet", or "multi-tube", not "{i}".')
        self._membrane_type = i.lower()

    @property
    def membrane_material(self):
        """
        [str] Can be any of the plastics ("PES", "PVDF", "PET", "PTFE")
        for any of the membrane types ("hollow fiber", "flat sheet", "multi-tube"),
        or "sintered steel" for "flat sheet",
        or "ceramic" for "multi-tube".
        """
        return self._membrane_material
    @membrane_material.setter
    def membrane_material(self, i):
        plastics = ('PES', 'PVDF', 'PET', 'PTFE')
        if i.upper() in plastics:
            self._membrane_material = i.upper()
        elif i.lower() in ('sintered steel', 'ceramic'):
            self._membrane_material = i.lower()
        else:
            raise ValueError(f'`membrane_material` can only be plastics materials '
                             f'{plastics}, "sintered steel", or "ceramic", not {i}.')


    ### Reactor/membrane tank ###
    @property
    def AF(self):
        """[:class:`~.FilterTank`] Anaerobic filter tank."""
        if self.reactor_type == 'CSTR':
            return None
        return self._AF

    @property
    def AeF(self):
        """[:class:`~.FilterTank`] Aerobic filter tank."""
        if not self.include_aerobic_filter:
            return None
        return self._AeF

    @property
    def N_train(self):
        """[int] Number of treatment train."""
        if not hasattr(self, '_N_train'):
            return self._N_train_min
        return self._N_train
    @N_train.setter
    def N_train(self, i):
        self._N_train = math.ceil(i)

    @property
    def cas_per_tank_spare(self):
        """[int] Number of spare cassettes per train."""
        return self._cas_per_tank_spare
    @cas_per_tank_spare.setter
    def cas_per_tank_spare(self, i):
        self._cas_per_tank_spare = math.ceil(i)

    @property
    def mod_per_cas_range(self):
        """
        [tuple] Range (min, max) of the number of membrane modules per cassette
        for the current membrane type.
        """
        return self._mod_per_cas_range
    @mod_per_cas_range.setter
    def mod_per_cas_range(self, i):
        self._mod_per_cas_range[self.membrane_type] = \
            tuple(math.floor(i[0]), math.floor(i[1]))

    @property
    def mod_per_cas(self):
        """
        [float] Number of membrane modules per cassette for the current membrane type.
        """
        return self._mod_per_cas or self._mod_per_cas_range[self.membrane_type][0]

    @property
    def cas_per_tank_range(self):
        """
        [tuple] Range (min, max) of the number of membrane cassette per tank
        (same for all membrane types).
        """
        return self._cas_per_tank_range
    @cas_per_tank_range.setter
    def cas_per_tank_range(self, i):
        self._cas_per_tank_range = tuple(math.floor(i[0]), math.floor(i[1]))

    @property
    def cas_per_tank(self):
        """
        [float] Number of membrane cassettes per tank for the current membrane type.
        """
        return self._cas_per_tank or self._cas_per_tank_range[0]

    @property
    def N_mod_tot(self):
        """[int] Total number of memberane modules."""
        return self.N_train * self.cas_per_tank * self.mod_per_cas

    @property
    def mod_surface_area(self):
        """
        [float] Surface area of the membrane for the current membrane type, [m2/module].
        Note that one module is one sheet for plat sheet and one tube for multi-tube.
        """
        return self._mod_surface_area[self.membrane_type]
    @mod_surface_area.setter
    def mod_surface_area(self, i):
        self._mod_surface_area[self.membrane_type] = float(i)

    @property
    def L_CSTR(self):
        """[float] Length of the CSTR tank, [ft]."""
        if self.reactor_type == 'AF':
            return 0
        return self._inf.F_vol/_ft3_to_m3*self.HRT/(self.N_train*self.W_tank*self.D_tank)

    @property
    def L_membrane_tank(self):
        """[float] Length of the membrane tank, [ft]."""
        return math.ceil((self.cas_per_tank+self.cas_per_tank_spare)*3.4)

    @property
    def W_tank(self):
        """[float] Width of the reactor/membrane tank (same value), [ft]."""
        return self._W_tank
    @W_tank.setter
    def W_tank(self, i):
        self._W_tank = float(i)

    @property
    def D_tank(self):
        """[float] Depth of the reactor/membrane tank (same value), [ft]."""
        return self._D_tank
    @D_tank.setter
    def D_tank(self, i):
        self._D_tank = float(i)

    @property
    def W_dist(self):
        """[float] Width of the distribution channel, [ft]."""
        return self._W_dist
    @W_dist.setter
    def W_dist(self, i):
        self._W_dist = float(i)

    @property
    def W_eff(self):
        """[float] Width of the effluent channel, [ft]."""
        return self._W_eff
    @W_eff.setter
    def W_eff(self, i):
        self._W_eff = float(i)

    @property
    def t_wall(self):
        """
        [float] Concrete wall thickness, [ft].
        Minimum of 1 ft with 1 in added for every ft of depth over 12 ft.
        """
        return 1 + max(self.D_tank-12, 0)/12

    @property
    def t_slab(self):
        """
        [float] Concrete slab thickness, [ft].
        2 in thicker than the wall thickness.
        """
        return self.t_wall+2/12

    @property
    def V_tot(self):
        """[float] Total volume of the unit, [ft3]."""
        return  self.D_tank*self.W_tank*self.L_CSTR*self.N_train

    @property
    def OLR(self):
        """[float] Organic loading rate, [kg COD/m3/hr]."""
        return  self.compute_COD(self.ins[0])*self.ins[0].F_vol/(self.V_tot*_ft3_to_m3)


    ### Pump/blower ###
    @property
    def N_blower(self):
        """
        [int] Number of blowers needed for gas sparging
        (not needed for some designs).
        Note that this is not used in costing
        (the cost is estimated based on the total sparging gas need).
        """
        if not self.add_GAC and self.membrane_configuration=='submerged':
            return self._N_blower
        return 0

    @property
    def N_degasser(self):
        """
        [int] Number of degassing membrane needed for dissolved biogas removal
        (optional).
        """
        if self.include_degassing_membrane:
            return math.ceil(self.Q_cmd/24/30) # assume each can hand 30 m3/d of influent
        return 0

    @property
    def W_PB(self):
        """[float] Width of the pump building, [ft]."""
        if self.membrane_configuration == 'submerged':
            N = self.cas_per_tank
        else: # cross-flow
            N = math.ceil(self.L_CSTR/((1+8/12)+(3+4/12)))

        if 0 <= N <= 10:
            W_PB = 27 + 4/12
        elif 11 <= N <= 16:
            W_PB = 29 + 6/12
        elif 17 <= N <= 22:
            W_PB = 31 + 8/12
        elif 23 <= N <= 28:
            W_PB = 35
        elif N >= 29:
            W_PB = 38 + 4/12
        else:
            W_PB = 0

        return W_PB

    @property
    def L_BB(self):
        """[float] Length of the blower building, [ft]."""
        if self.membrane_configuration == 'submerged':
            return (69+6/12) if self.cas_per_tank<=18 else (76 + 8/12)
        return 0

    @property
    def W_BB(self):
        """[float] Width of the blower building, [ft]."""
        if self.membrane_configuration == 'submerged':
            return (18+8/12) if self.cas_per_tank<=18 else 22
        return 0


    ### Wet well (submerged only) ###
    @property
    def L_well(self):
        """
        [float] Length of the wet well, [ft].
        Only needed for submerged configuration.
        """
        return self._L_well if self.membrane_configuration == 'submerged' else 0
    @L_well.setter
    def L_well(self, i):
        self._L_well = float(i)

    @property
    def W_well(self):
        """
        [float] Width of the wet well, [ft].
        Only needed for submerged configuration.
        """
        return self._W_well if self.membrane_configuration == 'submerged' else 0
    @W_well.setter
    def W_well(self, i):
        self._W_well = float(i)

    @property
    def D_well(self):
        """
        [float] Depth of the wet well, [ft].
        Only needed for submerged configuration.
        """
        return self._D_well if self.membrane_configuration == 'submerged' else 0
    @D_well.setter
    def D_well(self, i):
        self._D_well = float(i)


    ### Excavation ###
    @property
    def excav_slope(self):
        """[float] Slope for excavation (horizontal/vertical)."""
        return self._excav_slope
    @excav_slope.setter
    def excav_slope(self, i):
        self._excav_slope = float(i)

    @property
    def constr_access(self):
        """[float] Extra room for construction access, [ft]."""
        return self._constr_access
    @constr_access.setter
    def constr_access(self, i):
        self._constr_access = float(i)


    ### Operation-related parameters ###
    @property
    def Q_mgd(self):
        """
        [float] Influent volumetric flow rate in million gallon per day, [mgd].
        """
        return self.ins[0].F_vol*_m3_to_gal*24/1e6

    @property
    def Q_gpm(self):
        """[float] Influent volumetric flow rate in gallon per minute, [gpm]."""
        return self.Q_mgd*1e6/24/60

    @property
    def Q_cmd(self):
        """
        [float] Influent volumetric flow rate in cubic meter per day, [cmd].
        """
        return self.Q_mgd *1e6/_m3_to_gal # [m3/day]

    @property
    def Q_cfs(self):
        """[float] Influent volumetric flow rate in cubic feet per second, [cfs]."""
        return self.Q_mgd*1e6/24/60/60/_ft3_to_gal

    @property
    def HRT(self):
        """
        [float] Hydraulic retention time, [hr].
        """
        return self._HRT
    @HRT.setter
    def HRT(self, i):
        self._HRT = float(i)

    @property
    def recir_ratio(self):
        """
        [float] Internal recirculation ratio, will be updated in simulation
        if the originally set ratio is not adequate for the desired flow
        required by GAC (if applicable).
        """
        return self._recir_ratio
    @recir_ratio.setter
    def recir_ratio(self, i):
        self._recir_ratio = float(i)

    @property
    def J_max(self):
        """[float] Maximum membrane flux, [L/m2/hr]."""
        return self._J_max
    @J_max.setter
    def J_max(self, i):
        self._J_max = float(i)

    @property
    def J(self):
        """[float] Membrane flux, [L/m2/hr]."""
        # Based on the flux of one train being offline
        SA = (self.N_train-1) * self.cas_per_tank * self.mod_per_cas * self.mod_surface_area
        return self._inf.F_vol*1e3/SA # 1e3 is conversion from m3 to L

    @property
    def TMP_anaerobic(self):
        """[float] Transmembrane pressure in the anaerobic reactor, [psi]."""
        return self._TMP_dct[self.membrane_configuration]
    @TMP_anaerobic.setter
    def TMP_anaerobic(self, i):
        self._TMP_dct[self.membrane_configuration] = float(i)

    @property
    def TMP_aerobic(self):
        """
        [float] Transmembrane pressure in the aerobic filter, [psi].
        Defaulted to half of the reactor TMP.
        """
        if not self._include_aerobic_filter:
            return 0.
        else:
            return self._TMP_aerobic or self._TMP_dct[self.membrane_configuration]/2
    @TMP_aerobic.setter
    def TMP_aerobic(self, i):
        self._TMP_aerobic = float(i)

    @property
    def SGD(self):
        """[float] Specific gas demand, [m3 gas/m2 membrane area/h]."""
        return self._SGD
    @SGD.setter
    def SGD(self, i):
        self._SGD = float(i)

    @property
    def AFF(self):
        """
        [float] Air flow fraction, used in air pipe costing.
        The default value is calculated as STE/6
        (STE stands for standard oxygen transfer efficiency, and default STE is 20).
        If using different STE value, AFF should be 1 if STE/6<1
        and 3.33 if STE/6>1.
        """
        return self._AFF
    @AFF.setter
    def AFF(self, i):
        self._AFF = float(i)

    @property
    def v_cross_flow(self):
        """
        [float] Cross-flow velocity, [m/s].
        """
        return self._v_cross_flow if self.membrane_configuration=='cross-flow' else 0
    @v_cross_flow.setter
    def v_cross_flow(self, i):
        self._v_cross_flow = float(i)

    @property
    def v_GAC(self):
        """
        [float] Upflow velocity for GAC bed expansion, [m/hr].
        """
        return self._v_GAC if self.add_GAC==True else 0
    @v_GAC.setter
    def v_GAC(self, i):
        self._v_GAC = float(i)

    @property
    def biodegradability(self):
        """
        [float of dict] Biodegradability of chemicals,
        when shown as a float, all biodegradable chemicals are assumed to have
        the same degradability.
        """
        return self._biodegradability
    @biodegradability.setter
    def biodegradability(self, i):
        if not isinstance(i, dict):
            if not 0<=i<=1:
                raise ValueError('`biodegradability` should be within [0, 1], '
                                 f'the input value {i} is outside the range.')
            self._biodegradability = dict.fromkeys(self.chemicals.IDs, i)
        else:
            for k, v in i.items():
                if not 0<=v<=1:
                    raise ValueError('`biodegradability` should be within [0, 1], '
                                     f'the input value for chemical "{k}" is '
                                     'outside the range.')
            self._biodegradability = dict.fromkeys(self.chemicals.IDs, i).update(i)
        self._refresh_rxns()

    @property
    def sludge_conc(self):
        """Concentration of biomass ("WWTsludge") in the waste sludge, [g/L]."""
        return self._sludge_conc
    @sludge_conc.setter
    def sludge_conc(self, i):
        self._sludge_conc = i

    @property
    def i_rm(self):
        """[:class:`np.array`] Removal of each chemical in this reactor."""
        return self._i_rm

    @property
    def split(self):
        """Component-wise split to the treated water."""
        return self._split
    @split.setter
    def split(self, i):
        self._split = i
        self._isplit = self.chemicals.isplit(i, order=None)

    @property
    def biogas_rxns(self):
        """
        [:class:`tmo.ParallelReaction`] Biogas production reactions.
        """
        return self._biogas_rxns

    @property
    def growth_rxns(self):
        """
        [:class:`tmo.ParallelReaction`] Biomass (WWTsludge) growth reactions.
        """
        return self._growth_rxns

    @property
    def organic_rm(self):
        """[float] Overall organic (COD) removal rate."""
        Qi, Qe = self._inf.F_vol, self.outs[1].F_vol
        Si, Se = self.compute_COD(self._inf), self.compute_COD(self.outs[1])
        return 1 - Qe*Se/(Qi*Si)


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from ._unit import Unit
from .utils import piping

__all__ = ('temporary_connection', 'TemporaryUnit')

temporary_units_dump = []

def temporary_connection(source, sink):
    upstream = source.outs[0]
    downstream = sink.ins[0]
    temporary_stream = piping.TemporaryStream()
    if isinstance(upstream.sink, TemporarySource):
        upstream.sink.outs.append(temporary_stream)
    else:
        stream = piping.TemporaryStream()
        old_connection = upstream.get_connection()
        sink = upstream.sink
        if sink: upstream.sink.ins.replace(upstream, stream)
        TemporarySource(upstream, [stream, temporary_stream], old_connection)
    if isinstance(downstream.source, TemporarySink):
        downstream.source.ins.append(temporary_stream)
    else:
        stream = piping.TemporaryStream()
        old_connection = downstream.get_connection()
        downstream.sink.ins.replace(downstream, stream)
        TemporarySink([downstream, temporary_stream], stream, old_connection)

class TemporaryUnit:
    __slots__ = ('ID', '_ID', 'ins', 'outs', '_ins', '_outs', 'old_connection')
    auxiliary_units = ()
    def __init__(self, ins, outs, old_connection):
        temporary_units_dump.append(self)
        self.ID = self._ID = 'TU' + str(len(temporary_units_dump))
        self.old_connection = old_connection
        self.ins = self._ins = piping.Inlets(
            self, self._N_ins, ins, None, self._ins_size_is_fixed, 5
        )
        self.outs = self._outs = piping.Outlets(
            self, self._N_outs, outs, None, self._outs_size_is_fixed, 5
        )
        
    neighborhood = Unit.neighborhood
    get_downstream_units = Unit.get_downstream_units
    get_upstream_units = Unit.get_upstream_units
    _add_upstream_neighbors_to_set = Unit._add_upstream_neighbors_to_set
    _add_downstream_neighbors_to_set = Unit._add_downstream_neighbors_to_set
    __repr__ = Unit.__repr__
    __str__ = Unit.__str__


class TemporarySource(TemporaryUnit):
    __slots__ = ()
    _N_ins = 1
    _N_outs = 2
    _ins_size_is_fixed = True
    _outs_size_is_fixed = False
    

class TemporarySink(TemporaryUnit):
    __slots__ = ()
    _N_ins = 2
    _N_outs = 1
    _ins_size_is_fixed = False
    _outs_size_is_fixed = True
    
    


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import numpy as np
from .. import Unit
from .design_tools import size_batch
from .decorators import cost
import flexsolve as flx
from math import ceil

__all__ = ('BatchCrystallizer',)

crystallizer_material_factors = {
    'Carbon steel': 1.0,
    'Rubber lined carbon steel': 1.48,
    'Stainless steel': 3.43,
}

@cost('Crystallizer volume', 'Crystallizer',
      CE=444., S=0.003785411784, # Originally 1 gal
      BM=2.0, N='Number of crystallizers',
      f=lambda S: 222.4 * S**0.71 + 35150)
class BatchCrystallizer(Unit):
    """
    Abstract unit operation for atmospheric batch crystallization. Subclasses
    must implement the mass and energy balance through the `_run` method.
    
    Parameters
    ----------
    ins : 
        Inlet.
    outs : 
        Effluent.
    tau : float
        Reaction time [hr].
    N : int, optional
        Number of vessels. Either N or V must be given.
    V : float, optional
        Target volume of vessels [m^3].
    T=305.15 : float
        Operating temperature [K].
    Nmin=2 : int
        Minimum number of vessels.
    Nmax=36: int
        Maximum number of vessels.  
    kW : float, optional
        Electricity usage per volume in kW/gal. Defaults to 0.00746, a 
        heuristic value for suspension of solids.
    
    Notes
    -----
    The cost correlation for the crystallizer taken directly from the 
    Matches engineering firm website, which provides educational cost
    correlation tools for a variety of equipment [1]_. The default
    value of kW was retrieved from  [2]_.
    
    References
    ----------
    .. [1] Matches; certified engineering company at Edmond, Oklahoma.
    http://www.matche.com/equipcost/Crystallizer.html
    
    .. [2] Seider, W. D., Lewin,  D. R., Seader, J. D., Widagdo, S., Gani,
        R., & Ng, M. K. (2017). Product and Process Design Principles. Wiley.
        Cost Accounting and Capital Cost Estimation (Chapter 16)
    
    """
    _units = {'Crystallizer volume': 'm3',
              'Batch time': 'hr',
              'Loading time': 'hr'}
    
    _N_ins = _N_outs = 1
    
    #: [float] Cleaning and unloading time (hr).
    tau_0 = 1
    
    #: [float] Fraction of filled tank to total tank volume.
    V_wf = 0.9
    
    def _get_design_info(self):
        return (('Cleaning and unloading time', self.tau_0, 'hr'),
                ('Working volume fraction', self.V_wf, ''))
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None,
                 tau=None, N=None, V=None, T=305.15,
                 Nmin=2, Nmax=36, vessel_material='Carbon steel',
                 kW=0.00746):
        Unit.__init__(self, ID, ins, outs, thermo)
        self._N = N; self._V = V
        
        #: [float] Reaction time [hr].
        self.tau = tau
        
        #: [int] Number of vessels
        if N: self.N = N
        
        #: [float] Target volume of a vessel
        if V: self.V = V
        
        #: [float] Operating temperature of reactor [K].
        self.T = T
        
        #: [int] Minimum number of vessels
        self.Nmin = Nmin
        
        #: [int] Maximum number of vessels
        self.Nmax = Nmax
        
        #: [float] Electricity usage per volume in kW/gal
        self.kW = kW
        
    def _setup(self):
        super()._setup()
        effluent, = self.outs
        effluent.T = self.T
        effluent.P = 101325
    
    def _run(self):
        effluent, = self.outs
        effluent.copy_like(self.ins[0])
        effluent.T = self.T
        effluent.P = 101325
    
    @property
    def effluent(self):
        return self.outs[0]
        
    @property
    def N(self):
        """[int] Number of reactor."""
        return self._N
    @N.setter
    def N(self, N):
        if N <= 1:
            raise ValueError(f"number of crystallizers must be greater than 1, value {N} is infeasible")
        assert not self._V, 'cannot specify both crystallizer volume and number of crystallizers'
        self._N = ceil(N)

    @property
    def V(self):
        """[float] Crystallizer volume."""
        return self._V
    @V.setter
    def V(self, V):
        if V <= 1:
            raise ValueError(f"crystallizer volume must be greater than 1, value {V} is infeasible")
        assert not self._N, 'cannot specify both crystallizer volume and number of crystallizers'
        self._V = V

    @property
    def tau(self):
        return self._tau
    @tau.setter
    def tau(self, tau):
        self._tau = tau
        
    def _design(self):
        effluent = self.effluent
        v_0 = effluent.F_vol
        tau = self._tau
        tau_0 = self.tau_0
        V_wf = self.V_wf
        Design = self.design_results
        if self.V:
            f = lambda N: v_0 / N / V_wf * (tau + tau_0) / (1 - 1 / N) - self.V
            N = flx.IQ_interpolation(f, self.Nmin, self.Nmax,
                                     xtol=0.01, ytol=0.5, checkbounds=False)
            N = ceil(N)
        else:
            N = self._N
        dct = size_batch(v_0, tau, tau_0, N, V_wf)
        Design['Crystallizer volume'] = volume = dct.pop('Reactor volume')
        Design.update(dct)
        Design['Number of crystallizers'] = N
        self.add_heat_utility(self.Hnet, self.T)
        self.add_power_utility(self.kW * V_wf * volume * N)
        
    


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import biosteam as bst
from thermosteam import Stream

__all__ = (
    'streams_from_units',
    'process_streams',
    'streams_from_path', 
    'feeds',
    'products',
    'inlets',
    'outlets',
    'filter_out_missing_streams',
    'sort_feeds_big_to_small',
    'feeds_from_units',
    'products_from_units',
    'get_inlet_origin',
    'get_streams_from_context_level',
    'get_fresh_process_water_streams',
    'FreeProductStreams',
)

Stream.feed_priorities = {}

def inlets(units):
    return set(sum([i._ins for i in units], []))

def outlets(units):
    return set(sum([i._outs for i in units], []))

def streams_from_units(units):
    return set(sum([i._ins + i._outs for i in units], []))

def streams_from_path(path):
    isa = isinstance
    streams = set()
    System = bst.System
    Unit = bst.Unit
    for i in path:
        if isa(i, System):
            streams.update(i.streams)
        elif isa(i, Unit):
            streams.update(i._ins + i._outs)
    return streams

def get_inlet_origin(inlet):
    source = inlet.source
    while source:
        if len(source.ins) == len(source.outs) == 1 and 'processing' not in source.line.lower():
            inlet = source.ins[0]
        elif isinstance(source, bst.HXprocess):
            index = source.outs.index(inlet)
            inlet = source.ins[index]
        else:
            break
        source = inlet.source
    return inlet
        

def process_streams(streams):
    return {i for i in streams if i._source and i._sink}

def feeds(streams):
    return [s for s in streams if not s._source]

def products(streams):
    return [s for s in streams if not s._sink]

def filter_out_missing_streams(streams):
    streams.intersection_update([i for i in streams if i])

def sort_feeds_big_to_small(feeds):
    if feeds:
        feed_priorities = Stream.feed_priorities
        def feed_priority(feed):
            if feed in feed_priorities:
                return feed_priorities[feed]
            elif feed:
                return 1. - feed.F_mass / F_mass_max if F_mass_max else 1.
            else:
                return 2.
        F_mass_max = max([i.F_mass for i in feeds])
        feeds.sort(key=feed_priority)

def feeds_from_units(units):
    unit_set = set(units)
    return sum([[i for i in u._ins if i._source not in unit_set]
                 for u in units], [])

def products_from_units(units):
    unit_set = set(units)
    return sum([[i for i in u._outs if i._sink not in unit_set]
                 for u in units], [])

def get_feed_priority(stream):
    if stream.isfeed():
        return Stream.feed_priorities.get(stream)
    else:
        raise RuntimeError(f"stream '{stream}' is not a feed")

def set_feed_priority(stream, value):
    if stream.isfeed():
        Stream.feed_priorities[stream] = value
    else:
        raise RuntimeError(f"stream '{stream}' is not a feed")
    
Stream.get_feed_priority = get_feed_priority
Stream.set_feed_priority = set_feed_priority

def get_streams_from_context_level(level=None):
    """
    Return streams created in given context level.
    
    Parameters
    ----------
    level : int, optional
        If given, only filter through streams created in the given context 
        level. For example, use:
        * 0: to filter within the outer-most context level.
        * -1: to filter within the current context level.
        * -2: to filter within the outer context level.
        
    
    Examples
    --------
    Get all waste streams created (as registerd in the flowsheet):
        
    >>> import biosteam as bst
    >>> bst.main_flowsheet.clear()
    >>> bst.settings.set_thermo(['Water', 'Ethanol'], cache=True)
    >>> waste = bst.Stream('waste')
    >>> mixer = bst.Mixer('mixer', outs=waste)
    >>> streams = bst.get_streams_from_context_level()
    >>> assert len(streams) == 1 and streams[0] == waste
    
    Get waste streams created within a context:
        
    >>> with bst.System('sys') as sys:
    ...     new_waste = bst.Stream('new_waste')
    ...     tank = bst.MixTank('tank', outs=new_waste)
    ...     streams = bst.get_streams_from_context_level(level=0)
    ...     assert len(streams) == 1 and streams[0] == new_waste
    
    """
    if level is None:
        streams = list(bst.main_flowsheet.stream)
    else:
        units = bst.main_flowsheet.unit
        context_levels = units.context_levels
        N_levels = len(context_levels)
        if (level >= 0 and level >= N_levels or level < 0 and -level > N_levels):
            streams = list(bst.main_flowsheet.stream)
        else:
            units = context_levels[level]
            streams = [i for i in streams_from_units(units) if i]
    return streams

def get_fresh_process_water_streams(streams=None):
    """
    Return all feed water streams without a price.
    
    Parameters
    ----------
    streams : Iterable[Stream], optional
        Stream to filter through.
        
    Examples
    --------
    Get all fresh process water streams created (as registerd in the flowsheet):
        
    >>> import biosteam as bst
    >>> bst.main_flowsheet.clear()
    >>> bst.settings.set_thermo(['Water', 'Ethanol'], cache=True)
    >>> water = bst.Stream('water', Water=1)
    >>> mixer = bst.Mixer('mixer', ins=water)
    >>> streams = bst.get_fresh_process_water_streams()
    >>> assert len(streams) == 1 and streams[0] == water
    
    """    
    if streams is None: streams = bst.main_flowsheet.stream
    Facility = bst.Facility
    isa = isinstance
    return [
        i for i in streams
        if not i.price and i.isfeed() and not isa(i.sink, Facility) and has_only_water(i)
    ]

def has_only_water(stream):
    chemicals = stream.available_chemicals
    return len(chemicals) == 1 and chemicals[0].CAS == '7732-18-5'

class FreeProductStreams:
    __slots__ = ('all_streams', 'cache', 'LHV_combustible', 'wastewater_keys', 'combustible_keys')
    LHV_combustible_default = 500.
    wastewater_keys_default = {'wastewater', 'brine'}
    combustible_keys_default = {'to_boiler',}
    
    def __init__(self, all_streams, LHV_combustible=None, wastewater_keys=None, combustible_keys=None):
        self.all_streams = all_streams
        self.cache = {}
        self.wastewater_keys = self.wastewater_keys_default if wastewater_keys is None else wastewater_keys
        self.combustible_keys = self.combustible_keys_default if combustible_keys is None else combustible_keys
        self.LHV_combustible = self.LHV_combustible_default if LHV_combustible is None else LHV_combustible
       
    @property
    def streams(self):
        cache = self.cache
        if 'streams' in cache:
            return cache['streams']
        else:
            isa = isinstance
            cache['streams'] = streams = frozenset([
                i for i in self.all_streams if 
                not i.price and i.isproduct() and 'recycle' not in i._ID and not i.isempty()
                and not isa(i.source, bst.Facility)
            ])
        return streams
       
    @property
    def combustibles(self):
        cache = self.cache
        if 'cumbustibles' in cache: return cache['combustibles']
        LHV_combustible = self.LHV_combustible
        wastewater_keys = self.wastewater_keys
        combustible_keys = self.combustible_keys
        try:
            cache['combustibles'] = combustibles = frozenset([
                i for i in self.streams 
                if not any([j in i.ID.lower() for j in wastewater_keys])
                and (i.LHV / i.F_mass >= LHV_combustible)
            ] + [i for i in self.all_streams if any([j in i.ID.lower() for j in combustible_keys])])
        except:
            breakpoint()
        return combustibles
    
    @property
    def combustible_gases(self):
        cache = self.cache
        if 'combustible_gases' in cache: return cache['combustible_gases']
        combustibles = self.combustibles
        cache['combustible_gases'] = gas_combustibles = frozenset([i for i in combustibles if i.phase == 'g'])
        return gas_combustibles
    
    @property
    def combustible_slurries(self):
        cache = self.cache
        if 'combustible_slurries' in cache: return cache['combustible_slurries']
        combustible_gases = self.combustible_gases
        combustibles = self.combustibles
        cache['combustible_slurries'] = combustible_slurries = frozenset([i for i in combustibles if i not in combustible_gases])
        return combustible_slurries
    
    @property
    def noncombustibles(self):
        cache = self.cache
        if 'noncombustibles' in cache: return cache['noncombustibles']
        combustibles = self.combustibles
        cache['noncombustibles'] = noncombustibles = frozenset([i for i in self.streams if i not in combustibles])
        return noncombustibles
    
    @property
    def noncombustible_slurries(self):
        cache = self.cache
        if 'noncombustible_slurries' in cache: return cache['noncombustible_slurries']
        noncombustibles = self.noncombustibles
        cache['noncombustible_slurries'] = noncombustible_slurries = frozenset([i for i in noncombustibles if i.phase != 'g'])
        return noncombustible_slurries
        
    def __repr__(self):
        return f"{type(self).__name__}(all_streams={self.all_streams})"



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
This module includes classes and functions concerning Stream objects.
"""
from thermosteam.utils.decorators import registered_franchise
from thermosteam import Stream, MultiStream
from typing import NamedTuple
from warnings import warn
import biosteam as bst

__all__ = ('TemporaryStream', 'MissingStream', 'MockStream', 'Inlets',
           'Outlets', 'Sink', 'Source', 'InletPort', 'OutletPort', 'StreamPorts', 
           'Connection', 'as_stream', 'as_upstream', 'as_downstream', 
           'materialize_connections', 'ignore_docking_warnings',
           'IgnoreDockingWarnings', 'SuperpositionInlet', 'SuperpositionOutlet')

DOCKING_WARNINGS = True

def ignore_docking_warnings(f):
    def g(*args, **kwargs):
        global DOCKING_WARNINGS
        warn = DOCKING_WARNINGS
        DOCKING_WARNINGS = False
        try:
            return f(*args, **kwargs)
        finally:
            DOCKING_WARNINGS = warn
    g.__name__ = f.__name__
    return g

class IgnoreDockingWarnings:
    """
    Ignore docking warnings within a context by using an IgnoreDockingWarnings object.
    
    """
    __slots__ = ('original_value',)
    
    def __enter__(self): 
        global DOCKING_WARNINGS
        self.original_value = DOCKING_WARNINGS
        DOCKING_WARNINGS = False
        return 
    
    def __exit__(self, type, exception, traceback):
        global DOCKING_WARNINGS
        DOCKING_WARNINGS = self.original_value
        if exception: raise exception
        

# %% Utilities

def pipe_info(source, sink):
    """Return stream information header."""
    # First line
    if source is None:
        source = ''
    else:
        source = f' from {repr(source)}'
    if sink is None:
        sink = ''
    else:
        sink = f' to {repr(sink)}'
    return f"{source}{sink}"

def as_stream(stream):
    isa = isinstance
    if isa(stream, Stream):
        return stream
    elif isa(stream, str):
        return Stream(stream)
    elif stream is None:
        return MissingStream(None, None)
    
def as_upstream(stream, sink):
    stream = as_stream(stream)
    stream._sink = sink
    return stream
        
def as_downstream(stream, source):
    stream = as_stream(stream)
    stream._source = source
    return stream

def materialize_connections(streams):
    for stream in streams:
        if not stream: stream.materialize_connection()


# %% Special Stream objects    
    
class MissingStream:
    """
    Create a MissingStream object that acts as a dummy in Inlets and Outlets
    objects until replaced by an actual Stream object.
    """
    __slots__ = ('_source', '_sink')
    line = 'Stream'
    ID = 'missing stream'
    T = P = phase = None
    disconnect = Stream.disconnect
    disconnect_source = Stream.disconnect_source
    disconnect_sink = Stream.disconnect_sink
    isfeed = Stream.isfeed
    isproduct = Stream.isproduct
    source = Stream.source
    sink = Stream.sink
    
    def __init__(self, source=None, sink=None):
        self._source = source
        self._sink = sink
    
    def _get_tooltip_string(self, format, full):
        if format not in ('html', 'svg'): return ''
        return '(empty)'
            
    def materialize_connection(self, ID=""):
        """
        Disconnect this missing stream from any unit operations and 
        replace it with a material stream. 
        """
        source = self._source
        sink = self._sink
        if not (source or sink):
            raise RuntimeError("either a source or a sink is required to "
                               "materialize connection")
        material_stream = Stream(ID, thermo=(source or sink).thermo)
        if source: 
            try: 
                source._outs.replace(self, material_stream)
            except: # Must be an auxlet
                material_stream._source = source
        if sink: 
            try:
                sink._ins.replace(self, material_stream)
            except: # Must be an auxlet
                material_stream._sink = sink
        return material_stream
    
    def get_impact(self, key):
        return 0.
    
    def get_CF(self, key):
        return 0.
    
    def reset_cache(self):
        """Does nothing, MissingStream objects do not contain cache."""
    
    def get_data(self):
        return None
    
    def set_data(self, data): pass
    
    def get_total_flow(self, units):
        return 0.
    
    def get_flow(self, units, key=None):
        return 0.
    
    H = Hf = Hnet = LHV = HHV = Hvap = C = F_mol = F_mass = F_vol = cost = price = 0.
    
    def isempty(self):
        return True
    
    def empty(self): pass
    
    def __bool__(self):
        return False

    def __repr__(self):
        return f'<{type(self).__name__}>'
    
    def __str__(self):
        return self.ID
    
    def show(self):
        print(self._basic_info())

class TemporaryStream:
    """
    Create a TemporaryStream object that can serve in Inlets and Outlets
    objects to declare a specification dependency between units.
    """
    __slots__ = ('_source', '_sink')
    disconnect = Stream.disconnect
    disconnect_source = Stream.disconnect_source
    disconnect_sink = Stream.disconnect_sink
    __init__ = MissingStream.__init__
    source = Stream.source
    sink = Stream.sink
    H = Hf = Hnet = LHV = HHV = Hvap = C = F_mol = F_mass = F_vol = cost = price = 0.
    
    def __repr__(self):
        return f"<{type(self).__name__}>"


@registered_franchise(Stream)
class MockStream:
    """
    Create a MockStream object that acts as a dummy Stream object that is always
    empty. A MockStream behaves similar to a MissingStream, with the following 
    differences: they have an ID and are registered in the flowsheet as streams,
    they cannot serve as inlets or outlets and cannot be materialized, you can
    set life-cycle characterization factors.
    
    """
    __slots__ = ('_ID', 'characterization_factors', 'price')
    line = 'Stream'
    
    def __init__(self, ID):
        self._ID = ID
        self.characterization_factors = {}
        self.price = 0.
        self._register(ID)

    def _register(self, ID):
        replace_ticket_number = isinstance(ID, int)
        if replace_ticket_number: 
            Stream.ticket_numbers[Stream.ticket_name] = ID
        if ID == "" or replace_ticket_number: 
            registry = Stream.registry
            data = registry.data
            ID = Stream._take_ticket()
            while ID in data: ID = Stream._take_ticket()
            registry.register(ID, self)
        elif ID:
            Stream.registry.register_safely(ID, self) 
        else:
            self._ID = Stream._take_unregistered_ticket()
    
    source = sink = None
    
    @property
    def ID(self):
        """Unique identification (str). If set as '', it will choose a default ID."""
        return self._ID

    @ID.setter
    def ID(self, ID):
        Stream._register(self, ID)
    
    def isfeed(self):
        return False
    
    def isproduct(self):
        return False
    
    get_CF = Stream.get_CF
    set_CF = Stream.set_CF
    get_impact = MissingStream.get_impact
    reset_cache = MissingStream.reset_cache
    get_data = MissingStream.get_data
    set_data = MissingStream.set_data
    get_total_flow = MissingStream.get_total_flow
    get_flow = MissingStream.get_flow
    H = MissingStream.H
    Hf = MissingStream.Hf
    Hnet = MissingStream.Hnet
    LHV = MissingStream.LHV
    HHV = MissingStream.HHV
    Hvap = MissingStream.Hvap
    C = MissingStream.C
    F_mol = MissingStream.F_mol
    F_mass = MissingStream.F_mass
    F_vol = MissingStream.F_vol
    isempty = MissingStream.isempty
    empty = MissingStream.empty
    cost = MissingStream.cost
    
    __str__ = Stream.__str__
    __repr__ = Stream.__repr__

# %% Utilities

def n_missing(ub, N):
    if ub < N: raise RuntimeError(f"size of streams exceeds {ub}")
    return ub - N

# %% List objects for input and output streams

class StreamSequence:
    """
    Abstract class for a sequence of streams for a Unit object.
    
    Abstract methods:
        * _dock(self, stream) -> Stream
        * _redock(self, stream) -> Stream
        * _undock(self) -> None
        * _load_missing_stream(self)
    
    """
    __slots__ = ('_size', '_streams', '_fixed_size')
        
    def __init__(self, size, streams, thermo, fixed_size, stacklevel):
        self._size = size
        self._fixed_size = fixed_size
        dock = self._dock
        redock = self._redock
        if streams == ():
            self._streams = [dock(Stream(thermo=thermo)) for i in range(size)]
        else:
            isa = isinstance
            if fixed_size:
                self._initialize_missing_streams()
                if streams is not None:
                    if isa(streams, str):
                        self._streams[0] = dock(Stream(streams, thermo=thermo))
                    elif isa(streams, stream_types):
                        self._streams[0] = redock(streams, stacklevel)
                    else:
                        N = len(streams)
                        n_missing(size, N) # Make sure size is not too big
                        self._streams[:N] = [redock(i, stacklevel+1) if isa(i, stream_types)
                                             else dock(Stream(i, thermo=thermo)) for i in streams]
            elif streams is not None:
                if isa(streams, str):
                    self._streams = [dock(Stream(streams, thermo=thermo))]
                elif isa(streams, stream_types):
                    self._streams = [redock(streams, stacklevel)]
                else:
                    self._streams = loaded_streams = []
                    for i in streams:
                        if isa(i, stream_types):
                            s = redock(i, stacklevel)
                        elif i is None:
                            s = self._create_missing_stream()
                        else:
                            s = Stream(i, thermo=thermo)
                            dock(s)
                        loaded_streams.append(s)
            else:
                self._initialize_missing_streams()
        
    def _create_missing_stream(self):
        return MissingStream(None, None)
        
    def _create_N_missing_streams(self, N):
        return [self._create_missing_stream() for i in range(N)]
    
    def _initialize_missing_streams(self):
        #: All input streams
        self._streams = self._create_N_missing_streams(self._size)
        
    def __add__(self, other):
        return self._streams + other
    def __radd__(self, other):
        return other + self._streams
    
    # DO NOT DELETE: These should be implemented by child class
    # def _dock(self, stream): return stream
    # def _redock(self, stream, stacklevel): return stream
    # def _undock(self, stream): pass

    def _set_streams(self, slice, streams, stacklevel):
        streams = [self._as_stream(i) for i in streams]
        all_streams = self._streams
        for stream in all_streams[slice]: self._undock(stream)
        all_streams[slice] = streams
        stacklevel += 1
        for stream in all_streams: self._redock(stream, stacklevel)
        if self._fixed_size:
            size = self._size
            N_streams = len(all_streams)
            if N_streams < size:
                N_missing = n_missing(size, N_streams)
                if N_missing:
                    all_streams[N_streams: size] = self._create_N_missing_streams(N_missing)
       
    def _as_stream(self, stream):
        if stream is None:
            stream = self._create_missing_stream()
        elif not isinstance(stream, stream_types):
            raise TypeError(
                f"'{type(self).__name__}' object can only contain "
                f"'Stream' objects; not '{type(stream).__name__}'"
            )
        return stream
       
    @property
    def size(self):
        return self._streams.__len__()
    
    def __len__(self):
        return self._streams.__len__()
    
    def __bool__(self):
        return bool(self._streams)
    
    def _set_stream(self, int, stream, stacklevel):
        stream = self._as_stream(stream)
        self._undock(self._streams[int])
        self._streams[int] = self._redock(stream, stacklevel+1)
    
    def empty(self):
        for i in self._streams: self._undock(i)
        self._initialize_missing_streams()
    
    def insert(self, index, stream):
        if self._fixed_size: 
            raise RuntimeError(f"size of '{type(self).__name__}' object is fixed")
        self._undock(stream)
        self._dock(stream)
        self._streams.insert(index, stream)
    
    def append(self, stream):
        if self._fixed_size: 
            raise RuntimeError(f"size of '{type(self).__name__}' object is fixed")
        self._undock(stream)
        self._dock(stream)
        self._streams.append(stream)
    
    def extend(self, streams):
        if self._fixed_size: 
            raise RuntimeError(f"size of '{type(self).__name__}' object is fixed")
        for i in streams:
            self._undock(i)
            self._dock(i)
            self._streams.append(i)
    
    def replace(self, stream, other_stream):
        index = self.index(stream)
        self[index] = other_stream

    def index(self, stream):
        return self._streams.index(stream)

    def pop(self, index):
        streams = self._streams
        if self._fixed_size:
            stream = streams[index]
            missing_stream = self._create_missing_stream()
            self.replace(stream, missing_stream)
        else:
            stream = streams.pop(index)
        return stream

    def remove(self, stream):
        self._undock(stream)
        missing_stream = self._create_missing_stream()
        self.replace(stream, missing_stream)
        
    def clear(self):
        if self._fixed_size:
            self._initialize_missing_streams()
        else:
            for i in self._streams: self._undock(i)
            self._streams.clear()
    
    def reverse(self):
        self.streams.reverse()
    
    def __iter__(self):
        return iter(self._streams)
    
    def __getitem__(self, index):
        return self._streams[index]
    
    def __setitem__(self, index, item):
        isa = isinstance
        if isa(index, int):
            self._set_stream(index, item, 2)
        elif isa(index, slice):
            self._set_streams(index, item, 2)
        else:
            raise IndexError("Only intergers and slices are valid "
                             f"indices for '{type(self).__name__}' objects")
    
    def __repr__(self):
        return repr(self._streams)


class Inlets(StreamSequence):
    """Create an Inlets object which serves as input streams for a Unit object."""
    __slots__ = ('_sink', '_fixed_size')
    
    def __init__(self, sink, size, streams, thermo, fixed_size, stacklevel):
        self._sink = sink
        super().__init__(size, streams, thermo, fixed_size, stacklevel)
    
    @property
    def sink(self):
        return self._sink
    
    def _create_missing_stream(self):
        return MissingStream(None, self._sink)
    
    def _dock(self, stream): 
        stream._sink = self._sink
        return stream

    def _redock(self, stream, stacklevel): 
        sink = stream._sink
        if sink:
            ins = sink._ins
            if ins is not self:
                if stream in ins:
                    ins.remove(stream)
                    stream._sink = new_sink = self._sink
                    if (DOCKING_WARNINGS 
                        and sink._ID and new_sink._ID
                        and sink._ID != new_sink._ID):
                        warn(f"undocked inlet stream {stream} from unit {sink}; "
                             f"{stream} is now docked at {new_sink}", 
                             RuntimeWarning, stacklevel + 1)
                else:
                    stream._sink = self._sink
        else:
            stream._sink = self._sink
        return stream
    
    def _undock(self, stream): 
        stream._sink = None
    
        
class Outlets(StreamSequence):
    """Create an Outlets object which serves as output streams for a Unit object."""
    __slots__ = ('_source',)
    
    def __init__(self, source, size, streams, thermo, fixed_size, stacklevel):
        self._source = source
        super().__init__(size, streams, thermo, fixed_size, stacklevel)
    
    @property
    def source(self):
        return self._source
    
    def _create_missing_stream(self):
        return MissingStream(self._source, None)
    
    def _dock(self, stream): 
        stream._source = self._source
        return stream

    def _redock(self, stream, stacklevel): 
        source = stream._source
        if source:
            outs = source._outs
            if outs is not self:
                if stream in outs:
                    outs.remove(stream)
                    stream._source = new_source = self._source
                    if (DOCKING_WARNINGS 
                        and source._ID and new_source._ID
                        and source._ID != new_source._ID):
                        warn(f"undocked outlet stream {stream} from unit {source}; "
                             f"{stream} is now docked at {new_source}", 
                             RuntimeWarning, stacklevel + 1)
                else:
                    stream._source = self._source
        else:
            stream._source = self._source
        return stream
    
    def _undock(self, stream): 
        stream._source = None


# %% Sink and Source object for piping notation

class Sink:
    """
    Create a Sink object that connects a stream to a unit using piping
    notation:
    
    Parameters
    ----------
    stream : Stream
    index : int
        
    Examples
    --------
    First create a stream and a Mixer:
    
    .. code-block:: python
    
        >>> from biosteam import Stream, Mixer, settings
        >>> settings.set_thermo(['Water'])
        >>> stream = Stream('s1')
        >>> unit = Mixer('M1', outs=('out'))
    
    Sink objects are created using -pipe- notation:
        
    .. code-block:: python
    
        >>> stream-1
        <Sink: s1-1>
    
    Use pipe notation to create a sink and connect the stream:
    
    .. code-block:: python
    
        >>> stream-1-unit # The last unit is returned to continue piping; just ignore this
        <Mixer: M1>
        >>> unit.show()
        Mixer: M1
        ins...
        [0] missing stream
        [1] s1
            phase: 'l', T: 298.15 K, P: 101325 Pa
            flow: 0
        outs...
        [0] out
            phase: 'l', T: 298.15 K, P: 101325 Pa
            flow: 0
    
    """
    __slots__ = ('stream', 'index')
    def __init__(self, stream, index):
        self.stream = stream
        self.index = index

    # Forward pipping
    def __sub__(self, unit):
        unit.ins[self.index] = self.stream
        return unit
    
    # Backward pipping
    __pow__ = __sub__
    
    def __repr__(self):
        return '<' + type(self).__name__ + ': ' + self.stream.ID + '-' + str(self.index) + '>'


class Source:
    """
    Create a Source object that connects a stream to a unit using piping
    notation:
    
    Parameters
    ----------
    stream : Stream
    index : int
    
    Examples
    --------
    First create a stream and a Mixer:
    
    .. code-block:: python
    
        >>> from biosteam import Stream, Mixer, settings
        >>> settings.set_thermo(['Water'])
        >>> stream = Stream('s1')
        >>> unit = Mixer('M1')
    
    Source objects are created using -pipe- notation:
        
    .. code-block:: python
    
        >>> 1**stream
        <Source: 1-s1>
    
    Use -pipe- notation to create a source and connect the stream:
    
    .. code-block:: python
    
        >>> unit**0**stream # First unit is returned to continue backwards piping; just ignore this
        <Mixer: M1>
        >>> unit.show()
        Mixer: M1
        ins...
        [0] missing stream
        [1] missing stream
        outs...
        [0] s1
            phase: 'l', T: 298.15 K, P: 101325 Pa
            flow: 0
    
    """
    __slots__ = ('stream', 'index')
    def __init__(self, stream, index):
        self.stream = stream
        self.index = index

    # Forward pipping
    def __rsub__(self, unit):
        unit.outs[self.index] = self.stream
        return unit
    
    # Backward pipping
    __rpow__ = __rsub__
    
    def __repr__(self):
        return '<' + type(self).__name__ + ': ' + str(self.index) + '-' + self.stream.ID + '>'

# %% System pipping

class InletPort:
    __slots__ = ('sink', 'index')
    
    @classmethod
    def from_inlet(cls, inlet):
        sink = inlet.sink
        if not sink: raise ValueError(f'stream {inlet} is not an inlet to any unit')
        index = sink.ins.index(inlet)
        return cls(sink, index)
    
    def __init__(self, sink, index):
        self.sink = sink
        self.index = index
      
    def __eq__(self, other):
        return self.sink is other.sink and self.index == other.index  
      
    def _sorting_key(self):
        return (self.sink.ID[1:], self.sink.ID, self.index)
        
    def get_stream(self):
        return self.sink.ins[self.index]
    
    def set_stream(self, stream, stacklevel):
        self.sink.ins._set_stream(self.index, stream, stacklevel+1)
    
    def __str__(self):
        return f"{self.index}-{self.sink}"
    
    def __repr__(self):
        return f"{type(self).__name__}({self.sink}, {self.index})"


class OutletPort:
    __slots__ = ('source', 'index')
    
    @classmethod
    def from_outlet(cls, outlet):
        source = outlet.source
        if not source: raise ValueError(f'stream {outlet} is not an outlet to any unit')
        index = source.outs.index(outlet)
        return cls(source, index)
    
    def __init__(self, source, index):
        self.source = source
        self.index = index
    
    def __eq__(self, other):
        return self.source is other.source and self.index == other.index
    
    def _sorting_key(self):
        return (self.source.ID[1:], self.source.ID[0], self.index)
    
    def get_stream(self):
        return self.source.outs[self.index]
    
    def set_stream(self, stream, stacklevel):
        self.source.outs._set_stream(self.index, stream, stacklevel+1)
    
    def __str__(self):
        return f"{self.source}-{self.index}"
    
    def __repr__(self):
        return f"{type(self).__name__}({self.source}, {self.index})"


class StreamPorts:
    __slots__ = ('_ports',)
    
    @classmethod
    def from_inlets(cls, inlets, sort=None):
        return cls([InletPort.from_inlet(i) for i in inlets], sort)
    
    @classmethod
    def from_outlets(cls, outlets, sort=None):
        return cls([OutletPort.from_outlet(i) for i in outlets], sort)
    
    def __init__(self, ports, sort=None):
        if sort: ports = sorted(ports, key=lambda x: x._sorting_key())
        self._ports = tuple(ports)    
    
    def __bool__(self):
        return bool(self._ports)
        
    def __iter__(self):
        for i in self._ports: yield i.get_stream()
    
    def __len__(self):
        return len(self._ports)
    
    def __getitem__(self, index):
        if isinstance(index, slice):
            return self.__class__(self._ports[index])
        else:
            return self._ports[index].get_stream()
    
    def __setitem__(self, index, item):
        isa = isinstance
        if isa(index, int):
            self._set_stream(index, item, 2)
        elif isa(index, slice):
            self._set_streams(index, item, 2)
        else:
            raise IndexError("Only intergers and slices are valid "
                            f"indices for '{type(self).__name__}' objects")
          
    def _set_stream(self, int, stream, stacklevel):
        self._ports[int].set_stream(stream, stacklevel+1)
    
    def _set_streams(self, slice, streams, stacklevel):
        ports = self._ports[slice]
        stacklevel += 1
        if len(streams) == len(ports):
            for i, j in zip(ports, streams): i.set_stream(j, stacklevel)
        else:
            raise IndexError("number of inlets must match the size of slice")
    
    def __repr__ (self):
        ports = ', '.join([str(i) for i in self._ports])
        return f"[{ports}]"

# %% Auxiliary piping

def superposition_property(name):
    @property
    def p(self):
        return getattr(self.port.get_stream(), name)
    @p.setter
    def p(self, value):
        setattr(self.port.get_stream(), name, value)
        
    return p

def _superposition(cls, parent, port):
    excluded = set([*cls.__dict__, port, '_' + port, 'port'])
    for name in parent.__dict__:
        if name in excluded: continue
        setattr(cls, name, superposition_property(name))
    return cls

def superposition(parent, port):
    return lambda cls: _superposition(cls, parent, port)


@superposition(Stream, 'sink')
class SuperpositionInlet(Stream): # Both parent and auxiliary unit inlet.
    __slots__ = ()
    
    def __init__(self, port, sink=None):
        self.port = port
        self._sink = sink
      

@superposition(Stream, 'source')
class SuperpositionOutlet(Stream): # Both parent and auxiliary unit outlet.
    __slots__ = ()
    
    def __init__(self, port, source=None):
        self.port = port
        self._source = source
    
    
# %% Configuration bookkeeping

class Connection(NamedTuple):
    source: object
    source_index: int
    stream: object
    sink_index: int
    sink: object
    
    def reconnect(self):
        # Does not attempt to connect auxiliaries with owners (which should not be possible)
        source = self.source
        sink = self.sink
        if source:
            if (sink and getattr(sink, '_owner', None) is source):
                source.outs[self.source_index] = self.stream
        else:
            self.stream.disconnect_source()
        if sink:
            if not (source and getattr(source, '_owner', None) is sink):
                sink.ins[self.sink_index] = self.stream
        else:
            self.stream.disconnect_sink()


# %% General

stream_types = (Stream, MissingStream, TemporaryStream)

# %% Stream pipping
        
def __sub__(self, index):
    if isinstance(index, int):
        return Sink(self, index)
    elif isinstance(index, Stream):
        new = self.copy()
        new.separate_out(index)   
        return new
    else:
        return index.__rsub__(self)

def __rsub__(self, index):
    if isinstance(index, int):
        return Source(self, index)
    else:
        return index.__sub__(self)

def get_connection(self, junction=None):
    if junction is None: junction = True
    source = self._source
    sink = self._sink
    isa = isinstance
    if not junction and isa(source, bst.Junction):
        stream = source._ins[0]
        source = stream._source
    else:
        stream = self
    try:
        source_index = source._outs.index(stream) if source else None
    except ValueError: # Auxiliary streams are not in inlets nor outlets  
        try:
            auxname = sink.ID
            source_index = source._auxout_index[auxname]
        except:
            source_index = -1
    if not junction and isa(sink, bst.Junction):
        stream = sink._outs[0]
        sink = stream._sink
    else:
        stream = self
    try:
        sink_index = sink._ins.index(stream) if sink else None
    except ValueError: # Auxiliary streams are not in inlets nor outlets
        try:
            auxname = source.ID
            sink_index = sink._auxin_index[auxname]
        except:
            sink_index = -1
    return Connection(source, source_index, self, sink_index, sink)

TemporaryStream.get_connection = Stream.get_connection = \
MissingStream.get_connection = get_connection
basic_stream_info = lambda self: (f"{type(self).__name__}: {self.ID or ''}"
                                  f"{pipe_info(self._source, self._sink)}\n")
source_info = lambda self: f"{source}-{source.outs.index(self)}" if (source:=self.source) else self.ID
for cls in (MissingStream, Stream):
    cls.__pow__ = cls.__sub__ = __sub__  # Forward pipping
    cls.__rpow__ = cls.__rsub__ = __rsub__ # Backward pipping    
    cls._basic_info = basic_stream_info
    cls._source_info = source_info




# -*- coding: utf-8 -*-
"""
"""

__all__ = (
    'H_O2',
    'C_O2_L',
    'kLa',
    'P_at_kLa',
    'log_mean_driving_force',
)

from math import exp, log

kLa_coefficients = {
    # Garcia-Ochoa, F.; Gomez, E. Bioreactor Scale-up and Oxygen Transfer Rate
    # in Microbial Processes: An Overview. Biotechnology Advances 2009, 27 (2),
    # 153–176. https://doi.org/10.1016/j.biotechadv.2008.10.006.
    # Name: (a, b, c) # kLa = a * (P/V) ^ b * U^c 
    'Figueiredo & Calderbank': (0.026, 0.6, 0.8),
    "Van't Riet": (0.026, 0.4, 0.5),
}

def H_O2(T): 
    """
    Return Henry's law constant [mol O2 / kg / bar] for O2 given the temperature [K].
    
    Data from NIST Standard Reference Database 69: NIST Chemistry WebBook:
    https://webbook.nist.gov/cgi/cbook.cgi?ID=C7782447&Mask=10#Notes
    
    """
    k_H = 0.0013
    A = 1500
    dTinv = (1. / T - 1. / 298.15)
    return k_H * exp(A * dTinv)

def C_O2_L(T, P_O2): 
    """O2 concentration [mol / kg] in the liquid given the temperature [K] and oxygen
    partial pressure of O2 in the gas [bar]."""
    return P_O2 * H_O2(T)

def kLa(P, V, U, coefficients=None):
    """
    Return the lumped mass transfer coefficient and the
    mean bubble specific interfacial area (k_L*a; 1/s) given the
    gassed power input (P; W), the total volume (V; m3), and the
    gas superficial velocity in the reactor (m/s).
    
    Other Parameters
    ----------------
    coefficients : Iterable[float]|str, optional
        Name of author or an iterable in the form of a, b, c. The
        
    Notes
    ------
    The correlation is kLa = a * (P/V) ^ b * U ^ c.
        
    Correlation from:
    Van’t Riet, K. Review of Measuring Methods and Results in Nonviscous 
    Gas-Liquid Mass Transfer in Stirred Vessels. Ind. Eng. Chem. Proc. Des. 
    Dev. 1979, 18 (3), 357–364. https://doi.org/10.1021/i260071a001.


    """
    if coefficients is None:
        a, b, c = kLa_coefficients['Figueiredo & Calderbank']
    elif isinstance(coefficients, str):
        a, b, c = kLa_coefficients[coefficients]
    else:
        a, b, c = coefficients
    return a * (P / V) ** b * U ** c

def P_at_kLa(kLa, V, U, coefficients=None):
    """
    Return the gassed power input (P; W) given the lumped mass transfer 
    coefficient and the mean bubble specific interfacial area (k_L*a; 1/s),
    the total volume (V; m3), and the gas superficial velocity in the reactor (m/s).
    
    Other Parameters
    ----------------
    coefficients : Iterable[float]|str, optional
        Name of author or an iterable in the form of a, b, c. The
        
    Notes
    ------
    The correlation is kLa = a * (P/V) ^ b * U ^ c.
    
    Correlation from:
    Van’t Riet, K. Review of Measuring Methods and Results in Nonviscous 
    Gas-Liquid Mass Transfer in Stirred Vessels. Ind. Eng. Chem. Proc. Des. 
    Dev. 1979, 18 (3), 357–364. https://doi.org/10.1021/i260071a001.



    """
    if coefficients is None:
        a, b, c = kLa_coefficients['Figueiredo & Calderbank']
    elif isinstance(coefficients, str):
        a, b, c = kLa_coefficients[coefficients]
    else:
        a, b, c = coefficients
    return (kLa / (a * U ** c)) ** (1 / b) * V

def log_mean_driving_force(C_sat_out, C_sat_in, C_out, C_in=None):
    """
    Return the driving force for mass transfer. In small vessels (<1 m tall) 
    where both liquid concentration and saturation are almost constant, the simple
    form is adequate [1]_. In tall vessels, the log-mean driving force should be 
    used for more accuracy, since both the local concentration and the saturation
    concentration are different in the top and bottom of a bioreactor [1]_. 
    
    Parameters
    ----------
    C_sat_out : float
        Saturated concentration entering the bioreactor.
    C_sat_out : float
        Saturated concentration exiting the bioreactor.
    C_out : float, optional
        Outlet concentration. 
    C_in : float
        Inlet concentration. Defaults to the outlet concentration, which
        assumes perfect mixing and oxygen uptake rate is the same everywhere.
        
    References
    ----------
    [1] Benz, G. T. Bioreactor Design for Chemical Engineers. AICHE CEP 2011, 21–26.

    """
    if C_in is None: C_in = C_out # Assume perfect mixing and oxygen uptake rate is the same everywhere.
    dC_out = C_sat_out - C_out
    dC_in = C_sat_in - C_in
    if dC_out < 1e-9: # This is not theoretically possible, but may be an assumption
        dC_out = 1e-9 # Assume it near zero
    return (dC_out - dC_in) / log(dC_out / dC_in)


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import numpy as np
import pandas as pd
import thermosteam as tmo
import biosteam as bst
from .._heat_utility import HeatUtility

DataFrame = pd.DataFrame
ExcelWriter = pd.ExcelWriter

__all__ = ('stream_table', 'cost_table', 'unit_reaction_tables',
           'unit_result_tables', 'heat_utility_tables',
           'power_utility_table', 'tables_to_excel', 'voc_table',
           'lca_displacement_allocation_table', 
           'lca_inventory_table',
           'lca_property_allocation_factor_table',
           'lca_displacement_allocation_factor_table',
           'FOCTableBuilder')

def _stream_key(s): # pragma: no coverage
    num = s.ID[1:]
    if num.isnumeric(): return int(num)
    else: return -1

def _reformat(name):
    name = name.replace('_', ' ')
    if name.islower(): name= name.capitalize()
    return name

# %% Detailed TEA tables

class FOCTableBuilder:
    __slots__ = ('index', 'data', 'costs')
    
    def __init__(self):
        self.index = []
        self.data = []
        self.costs = []
        
    def entry(self, index, costs, notes='-'):
        self.index.append(index)
        self.data.append([notes, *costs])
        self.costs.append(costs)
    
    def table(self, names):
        data = self.data
        index = self.index.copy()
        index.append('Fixed operating cost (FOC)')
        data.append(("", *sum(self.costs)))
        return pd.DataFrame(np.array(data), 
                            index=index,
                            columns=(
                                'Notes',
                                *[i + '\n[MM$ / yr]' for i in names],
                            )
        )

# %% Multiple system tables

def voc_table(systems, product_IDs, system_names=None, unit='MT', with_products=False):
    # Not ready for users yet
    isa = isinstance
    if isa(systems, bst.System): systems = [systems]
    other_utilities_dct = {}
    other_byproducts_dct = {}
    prices = bst.stream_utility_prices
    factor = 1 / tmo.units_of_measure.convert(1, 'kg', unit)
    def getsubdct(dct, name):
        if name in dct:
            subdct = dct[name]
        else:
            dct[name] = subdct = {}
        return subdct
    
    for sys in systems:
        electricity_cost = sys.power_utility.cost * sys.operating_hours
        if electricity_cost > 0.: 
            dct = getsubdct(other_utilities_dct, 'Electricity')
            dct[sys] = (f"{bst.PowerUtility.price} $/kWh", electricity_cost)
        else:
            dct = getsubdct(other_byproducts_dct, 'Electricity production')
            dct[sys] = (f"{bst.PowerUtility.price} $/kWh", -electricity_cost)
        inlet_flows = sys.get_inlet_utility_flows()
        for name, flow in inlet_flows.items():
            dct = getsubdct(other_utilities_dct, name)
            price = prices[name]
            dct[sys] = (price * factor, price * flow)
        outlet_flows = sys.get_outlet_utility_flows()
        for name, flow in outlet_flows.items():
            dct = getsubdct(other_byproducts_dct, name)
            price = prices[name]
            dct[sys] = (price * factor, price * flow)
        
    def reformat(name):
        name = name.replace('_', ' ')
        if name.islower(): name= name.capitalize()
        return name
    
    feeds = sorted({i.ID for i in sum([i.feeds for i in systems], []) if sys.has_market_value(i)})
    coproducts = sorted({i.ID for i in sum([i.products for i in systems], []) if sys.has_market_value(i) and i.ID not in product_IDs})
    system_heat_utilities = [bst.HeatUtility.sum_by_agent(sys.heat_utilities) for sys in systems]
    other_utilities = sorted(other_utilities_dct)
    other_byproducts = sorted(other_byproducts_dct)
    heating_agents = sorted(set(sum([[i.agent.ID for i in hus if i.cost and i.flow * i.duty > 0. and abs(i.flow) > 1e-6] for hus in system_heat_utilities], [])))
    cooling_agents = sorted(set(sum([[i.agent.ID for i in hus if i.cost and i.flow * i.duty < 0. and abs(i.flow) > 1e-6] for hus in system_heat_utilities], [])))
    index = {j: i for (i, j) in enumerate(feeds + heating_agents + cooling_agents + other_utilities + coproducts + other_byproducts)}
    table_index = [*[('Raw materials', reformat(i)) for i in feeds],
                   *[('Heating utilities', reformat(i)) for i in heating_agents],
                   *[('Cooling utilities', reformat(i)) for i in cooling_agents],
                   *[('Other utilities', i) for i in other_utilities],
                   *[('By-products and credits', reformat(i)) for i in coproducts],
                   *[('By-products and credits', i) for i in other_byproducts]]
    table_index.append(('Variable operating cost', ''))
    if with_products:
        table_index.extend(
            [('Products', reformat(i)) for i in product_IDs]
        )
        for i, ID in enumerate(reversed(product_IDs)): index[ID] = -(i + 1)
    N_cols = len(systems) + 1
    N_rows = len(table_index)
    data = np.zeros([N_rows, N_cols], dtype=object)
    N_coproducts = len(coproducts) + len(other_byproducts)
    for col, sys in enumerate(systems):
        for stream in sys.feeds + sys.products:
            if stream.ID in product_IDs and not with_products: continue
            cost = sys.get_market_value(stream)
            if cost:
                ind = index[stream.ID]
                data[ind, 0] = stream.price * factor # USD / ton
                data[ind, col + 1] = cost / 1e6 # million USD / yr
        for hu in system_heat_utilities[col]:
            try:
                ind = index[hu.agent.ID]
            except:
                continue
            cost = sys.operating_hours * hu.cost
            price = ""
            if hu.agent.heat_transfer_price:
                price += f"{hu.agent.heat_transfer_price} USD/kJ"
            if hu.agent.regeneration_price:
                if price: price += ", "
                price += f"{hu.agent.regeneration_price} USD/kmol"
            data[ind, 0] = price 
            data[ind, col + 1] = cost / 1e6 # million USD / yr
        for sysdct in (other_byproducts_dct, other_utilities_dct):
            for i, dct in sysdct.items():
                if sys not in dct: continue
                price, cost = dct[sys]
                ind = index[i]
                data[ind, 0] = price
                data[ind, col + 1] = cost / 1e6 # million USD / yr
    N_consumed = N_rows - N_coproducts - 1
    N_products = len(product_IDs) if with_products else 0
    VOC_index = N_rows - N_products - 1
    data[VOC_index, 1:] = data[:N_consumed, 1:].sum(axis=0) - data[N_consumed:VOC_index, 1:].sum(axis=0)
    if system_names is None:
        system_names = [i.ID for i in systems]
    columns = [i + " [MM$/yr]" for i in system_names]
    return pd.DataFrame(data, 
                        index=pd.MultiIndex.from_tuples(table_index),
                        columns=(f'Price [$/{unit}]', *columns))

def lca_inventory_table(systems, key, items=(), system_names=None):
    isa = isinstance
    if items:
        item = items[0]
        if isa(item, list): 
            items = sum(items, [])
        elif isa(item, tuple): 
            items = sum(items, ())
    items = frozenset(items)
    if isa(systems, bst.System): systems = [systems]
    PowerUtility = bst.PowerUtility
    other_utilities = []
    other_byproducts = []
    process_items = []
    other_values = {}
    def set_value(name, sys, value):
        if name in other_values:
            dct = other_values[name]
        else:
            other_values[name] = dct = {}
        dct[sys] = value
        
    for sys in systems:
        electricity_consumption = sys.power_utility.rate * sys.operating_hours
        if electricity_consumption > 0.: 
            if 'Electricity [kWhr/yr]' not in other_utilities: other_utilities.append('Electricity [kWhr/yr]')
            cf = PowerUtility.get_CF(key, production=False)
            if cf:
                set_value('Electricity [kWhr/yr]', sys, electricity_consumption)
        elif electricity_consumption < 0.:
            if 'Electricity [kWhr/yr]' not in other_byproducts: other_byproducts.append('Electricity [kWhr/yr]')
            cf = PowerUtility.get_CF(key, consumption=False)
            if cf:
                set_value('Electricity [kWhr/yr]', sys, -electricity_consumption)
        try: process_impact_items = sys.process_impact_items[key]
        except: continue
        for item in process_impact_items:
            if item.name not in process_items: process_items.append(item.name)
            value = item.inventory()
            if item.basis != 'kg':
                value = f"{value} [{item.basis}/yr]"
            set_value(item.name, sys, value)
        
    def reformat(name):
        name = name.replace('_', ' ')
        if name.islower(): name= name.capitalize()
        return name
    
    feeds = sorted({i.ID for i in sum([i.feeds for i in systems], []) if key in i.characterization_factors})
    coproducts = sorted({i.ID for i in sum([i.products for i in systems], []) if key in i.characterization_factors or i in items})
    system_heat_utilities = [bst.HeatUtility.sum_by_agent(sys.heat_utilities) for sys in systems]
    input_heating_agents = sorted(set(sum([[i.agent.ID for i in hus if (key, i.agent.ID) in i.characterization_factors and i.flow * i.duty > 0. and i.flow > 1e-6] for hus in system_heat_utilities], [])))
    input_cooling_agents = sorted(set(sum([[i.agent.ID for i in hus if (key, i.agent.ID) in i.characterization_factors and i.flow * i.duty < 0. and i.flow > 1e-6] for hus in system_heat_utilities], [])))
    output_heating_agents = sorted(set(sum([[i.agent.ID for i in hus if (key, i.agent.ID) in i.characterization_factors and i.flow * i.duty > 0. and i.flow < -1e-6] for hus in system_heat_utilities], [])))
    output_cooling_agents = sorted(set(sum([[i.agent.ID for i in hus if (key, i.agent.ID) in i.characterization_factors and i.flow * i.duty < 0. and i.flow < -1e-6] for hus in system_heat_utilities], [])))
    index = {j: i for (i, j) in enumerate(feeds + input_heating_agents + input_cooling_agents + other_utilities + coproducts + other_byproducts + output_heating_agents + output_cooling_agents + process_items)}
    table_index = [*[('Inputs', reformat(i)) for i in feeds + input_heating_agents + input_cooling_agents + other_utilities],
                   *[('Outputs', reformat(i)) for i in coproducts + other_byproducts + output_heating_agents + output_cooling_agents],
                   *[(i, '') for i in process_items]]
    N_cols = len(systems)
    N_rows = len(table_index)
    data = np.zeros([N_rows, N_cols], dtype=object)
    for col, sys in enumerate(systems):
        for stream in sys.feeds + sys.products:
            try: ind = index[stream.ID]
            except: continue
            data[ind, col] = sys.get_mass_flow(stream)
        for hu in system_heat_utilities[col]:
            try: ind = index[hu.agent.ID]
            except: continue
            flow, units = hu.get_inventory()
            if flow:
                flow = sys.operating_hours * flow
                data[ind, col] = f"{flow} [{units}]"
        for i, subdct in other_values.items():
            if sys not in subdct: continue
            value = subdct[sys]
            ind = index[i]
            data[ind, col] = value
    if system_names is None and len(systems) == 1:
        columns = ["Inventory [kg/yr]"]
    else:
        if system_names is None:
            system_names = [i.ID for i in systems]
        sys_units = " inventory [kg/yr]"
        columns = [i + sys_units for i in system_names]
    return pd.DataFrame(data, 
                        index=pd.MultiIndex.from_tuples(table_index),
                        columns=columns)

def lca_displacement_allocation_table(systems, key, items, 
                                      item_name=None, system_names=None):
    # Not ready for users yet
    isa = isinstance
    if isa(systems, bst.System): systems = [systems]
    PowerUtility = bst.PowerUtility
    other_utilities = []
    other_byproducts = []
    process_inventory = []
    other_values = {}
    impact_units = str(bst.settings.get_impact_indicator_units(key))
    def set_value(name, sys, CF, value):
        if name in other_values:
            dct = other_values[name]
        else:
            other_values[name] = dct = {}
        dct[sys] = (CF, value)
        
    for sys in systems:
        electricity_consumption = sys.power_utility.rate * sys.operating_hours
        if electricity_consumption > 0.: 
            cf = PowerUtility.get_CF(key, production=False)
            if not cf: continue
            if 'Electricity' not in other_utilities: other_utilities.append('Electricity')
            set_value('Electricity', sys, f"{cf} {impact_units}/kWhr", electricity_consumption * cf)
        elif electricity_consumption < 0.:
            cf = PowerUtility.get_CF(key, consumption=False)
            if not cf: continue
            if 'Electricity' not in other_byproducts: other_byproducts.append('Electricity')
            set_value('Electricity', sys, f"{cf} {impact_units}/kWhr", - electricity_consumption * cf)
        try: process_impact_items = sys.process_impact_items[key]
        except: continue
        for item in process_impact_items:
            if item.name not in process_inventory: process_inventory.append(item.name)
            value = item.impact()
            CF = item.CF
            basis = item.basis
            if basis != 'kg':
                CF = f"{CF} [{impact_units}/{basis}"
                value = f"{value} [{basis}/yr]"
            set_value(item.name, sys, CF, value)
    
    if item_name is None: 
        item = items[0]
        if hasattr(item , 'ID'):
            item_name = item.ID.replace('_', ' ')
        else:
            item_name = item[0].ID.replace('_', ' ')
    feeds = sorted({i.ID for i in sum([i.feeds for i in systems], []) if key in i.characterization_factors})
    coproducts = sorted({i.ID for i in sum([i.products for i in systems], []) if key in i.characterization_factors})
    system_heat_utilities = [bst.HeatUtility.sum_by_agent(sys.heat_utilities) for sys in systems]
    input_heating_agents = sorted(set(sum([[i.agent.ID for i in hus if (key, i.agent.ID) in i.characterization_factors and i.flow * i.duty > 0. and i.flow > 1e-6] for hus in system_heat_utilities], [])))
    input_cooling_agents = sorted(set(sum([[i.agent.ID for i in hus if (key, i.agent.ID) in i.characterization_factors and i.flow * i.duty < 0. and i.flow > 1e-6] for hus in system_heat_utilities], [])))
    output_heating_agents = sorted(set(sum([[i.agent.ID for i in hus if (key, i.agent.ID) in i.characterization_factors and i.flow * i.duty > 0. and i.flow < -1e-6] for hus in system_heat_utilities], [])))
    output_cooling_agents = sorted(set(sum([[i.agent.ID for i in hus if (key, i.agent.ID) in i.characterization_factors and i.flow * i.duty < 0. and i.flow < -1e-6] for hus in system_heat_utilities], [])))
    inputs = [*feeds, *input_heating_agents, *input_cooling_agents, *other_utilities]
    outputs = [*coproducts, *other_byproducts, *output_heating_agents, *output_cooling_agents]
    keys = [*inputs, 'Total inputs', *outputs, 'Total outputs displaced', *process_inventory, 'Total']
    index = {j: i for (i, j) in enumerate(keys)}
    table_index = [*[('Inputs', _reformat(i)) for i in feeds + input_heating_agents + input_cooling_agents + other_utilities],
                   ('Total inputs', ''),
                   *[('Outputs displaced', _reformat(i)) for i in coproducts + other_byproducts + output_heating_agents + output_cooling_agents],
                   ('Total outputs displaced', ''),
                   *[("Process impacts", i) for i in process_inventory],
                   ("Total", '')]
    N_cols = len(systems) + 1
    N_rows = len(table_index)
    data = np.zeros([N_rows, N_cols], dtype=object)
    isa = isinstance
    for col, sys in enumerate(systems):
        item = items[col]
        if isa(item, bst.Stream): item = [item]
        item_flow = sum([sys.get_mass_flow(i) for i in item]) 
        for stream in sys.feeds + sys.products:
            try: 
                ind = index[stream.ID]
                data[ind, 0] = stream.characterization_factors[key]
            except: continue
            impact = sys.get_material_impact(stream, key)
            data[ind, col + 1] = impact / item_flow
        for hu in system_heat_utilities[col]:
            try: ind = index[hu.agent.ID]
            except: continue
            impact = sys.operating_hours * hu.get_impact(key)
            cf, basis = hu.characterization_factors[hu.agent.ID, key]
            data[ind, 0] = f"{cf} [{impact_units}{basis}]"
            data[ind, col + 1] = impact / item_flow
        for i, subdct in other_values.items():
            if sys not in subdct: continue
            cf, impact = subdct[sys]
            ind = index[i]
            data[ind, 0] = cf
            data[ind, col + 1] = impact / item_flow
    total_inputs_index = index['Total inputs']
    data[total_inputs_index, 1:] = total_inputs = data[:total_inputs_index, 1:].sum(axis=0)
    data[total_inputs_index, 0] = ''
    total_outputs_index = index['Total outputs displaced']
    data[total_outputs_index, 1:] = total_outputs = data[total_inputs_index + 1:total_outputs_index, 1:].sum(axis=0)
    data[total_outputs_index, 0] = ''
    data[-1, 1:] = total_inputs - total_outputs + data[total_outputs_index + 1:-1, 1:].sum(axis=0)
    data[-1, 0] = ''
    if system_names is None and len(systems) == 1:
        columns = [f"{key} [{impact_units}/kg*{item_name}]"]
    else:
        if system_names is None:
            system_names = [i.ID for i in systems]
        sys_units = f" {key} [{impact_units}/kg*{item_name}]"
        columns = [i + sys_units for i in system_names]
    return pd.DataFrame(data, 
                        index=pd.MultiIndex.from_tuples(table_index),
                        columns=(f'Characterization factor [{impact_units}/kg]', *columns))

def lca_property_allocation_factor_table(
        systems, property, units=None, system_names=None, groups=None,
    ):
    if groups is None: groups = {}
    system_allocation_factors = [i.get_property_allocation_factors(property, units, groups) for i in systems]
    table_index = sorted(set(sum([tuple(i) for i in system_allocation_factors], ())))
    index = {j: i for i, j in enumerate(table_index)}
    N_cols = len(systems)
    N_rows = len(table_index)
    data = np.zeros([N_rows, N_cols], dtype=object)
    for col, factors in enumerate(system_allocation_factors):
        for key, value in factors.items():
            data[index[key], col] = value
    if system_names is None and len(systems) == 1:
        columns = [f"{_reformat(property)} allocation factors"]
    else:
        if system_names is None:
            system_names = [i.ID for i in systems]
        sys_units = f" {property.replace('_', ' ')} allocation factors"
        columns = [i + sys_units for i in system_names]
    return pd.DataFrame(data, 
                        index=table_index,
                        columns=columns)

def lca_displacement_allocation_factor_table(
        systems, items, key, system_names=None, groups=None
    ):
    if groups is None: groups = {}
    system_allocation_factors = [i.get_displacement_allocation_factors(j, key, groups) for i, j in zip(systems, items)]
    table_index = sorted(set(sum([tuple(i) for i in system_allocation_factors], ())))
    index = {j: i for i, j in enumerate(table_index)}
    N_cols = len(systems)
    N_rows = len(table_index)
    data = np.zeros([N_rows, N_cols], dtype=object)
    for col, factors in enumerate(system_allocation_factors):
        for key, value in factors.items():
            data[index[key], col] = value
    if system_names is None and len(systems) == 1:
        columns = ["Displacement allocation factors"]
    else:
        if system_names is None:
            system_names = [i.ID for i in systems]
        sys_units = " displacement allocation factors"
        columns = [i + sys_units for i in system_names]
    return pd.DataFrame(data, 
                        index=table_index,
                        columns=columns)

def lca_allocation_method_results(
        systems, key, property_args, items, item_name 
    ):
    pass

# %% Helpful functions

def tables_to_excel(tables, writer, sheet='Sheet1', n_row=1, row_spacing=2): 
    """
    Save a list of tables as an excel file and return the row number at which
    another consecutive table would start.
    
    Parameters
    ----------
    tables : iterable[pandas.DataFrame]
        Tables to save to excel.
    writer : pandas.ExcelWritter
        Writer to manage data stream to excel.
    sheet : str
        Name of sheet to save data.
    n_row : int
        Row number to begin saving data.
    row_spacing : int
        Number of rows between tables.
    
    Returns
    -------
    n_row : int
        Row number for next table.
        
    """
    row_spacing += 1 # Account for Python index offset
    for t in tables:
        label = t.columns.name
        t.to_excel(writer, sheet, 
                   startrow=n_row, index_label=label)
        n_row += len(t.index) + row_spacing
    return n_row

# %% Units

def unit_reaction_tables(units):
    isa = isinstance
    tables = []
    rxntypes = (tmo.Reaction, tmo.ReactionSet)
    for u in units:
        all_reactions = {rxn for rxn in u.__dict__.values() if isa(rxn, rxntypes)}
        for rxn in tuple(all_reactions):
            if hasattr(rxn, '_parent'):
                if rxn._parent in all_reactions: all_reactions.discard(rxn)
            elif hasattr(rxn, '_parent_index'):
                parent, index = rxn._parent_index
                if parent in all_reactions: all_reactions.discard(rxn)
        for rxn in all_reactions:
            df = rxn.to_df()
            df.columns.name = '-'.join([u.ID, u.line])
            tables.append(df)
    return tables

def unit_result_tables(units,
                       include_utilities=False, 
                       include_total_cost=False,
                       include_installed_cost=False): 
    """
    Return a list of results tables for each unit type.

    Parameters
    ----------
    units : iterable[Unit]
        
    Returns
    -------
    tables : list[DataFrame]
    
    """
    units = sorted(units, key=(lambda u: u.line))
    
    # Organize units by units of measure:
    organized = {}
    for u in units:
        uom = (*u._units.keys(), u.line)
        if uom in organized: organized[uom].append(u)
        else: organized[uom] = [u]
    
    # Make a list of tables, keeping all results with same keys in one table
    tables = []
    key = lambda u: u.ID
    for all_units in organized.values():
        # First table with units of measures
        all_units = sorted(all_units, key=key)
        u, *units = all_units
        empty_heat_utilities = []
        key_hook = None
        if all([i.line == 'Heat Exchanger' for i in all_units]):
            key_hook = lambda key: ('Purchase cost', 'Heat Exchanger') if key[0]=='Purchase cost' else key
            all_heat_utilities = sum([i.heat_utilities for i in all_units], ())
            for i in set([i.agent for i in all_heat_utilities]):
                if not i: continue
                heat_utility = HeatUtility()
                heat_utility.load_agent(i)
                empty_heat_utilities.append(heat_utility)
        empty_heat_utilities = tuple(empty_heat_utilities)
        table = u.results(include_utilities=include_utilities,
                          include_total_cost=include_total_cost,
                          include_installed_cost=include_installed_cost,
                          include_zeros=False,
                          external_utilities=empty_heat_utilities,
                          key_hook=key_hook)
        if table is None: continue
        for u in units:
            table[u.ID] = u.results(with_units=False, 
                                    include_utilities=include_utilities,
                                    include_total_cost=include_total_cost,
                                    include_installed_cost=include_installed_cost,
                                    include_zeros=False,
                                    external_utilities=empty_heat_utilities,
                                    key_hook=key_hook)
        table.columns.name = (u.line, '')
        tables.append(table)
    return tables
    
def cost_table(tea): 
    """
    Return a cost table as a pandas DataFrame object.

    Parameters
    ----------
    units : iterable[Unit]
        
    Returns
    -------
    table : DataFrame

    """
    columns = ('Unit operation',
               'Purchase cost (10^6 USD)',
               'Utility cost (10^6 USD/yr)')
    units = sorted([i for i in tea.system.units if i._design or i._cost], key=lambda x: x.line)
    operating_days = tea.operating_days
    N_units = len(units)
    array = np.empty((N_units, 3), dtype=object)
    IDs = []
    types = array[0:, 0]
    C_cap = array[0:, 1]
    C_op = array[0:, 2]
    
    # Get data
    for i in range(N_units):
        unit = units[i]
        types[i] = unit.line
        C_cap[i] = unit.purchase_cost / 1e6
        C_op[i] = unit.utility_cost * operating_days * 24  / 1e6
        IDs.append(unit.ID)
    
    df = DataFrame(array, columns=columns, index=IDs)    
    if not tea.lang_factor:
        df['Installed cost (10^6 USD)'] = [u.installed_cost / 1e6 for u in units]
    
    return df

def heat_utility_tables(units): 
    """Return a list of utility tables for each heat utility source.
    
    Parameters
    ----------
    units : iterable[Unit]
        
    Returns
    -------
    tables : list[DataFrame]
        
    """
    # Sort heat utilities by unit type, then by utility Type
    units = sorted(units, key=(lambda u: type(u).__name__))
    source = {}
    heat_utils = []
    for u in units:
        hus = u.heat_utilities
        if not hus: continue
        for i in hus: source[i] = u
        heat_utils.extend(hus)

    # Organize heatutility by ID
    heat_utils_dict = {}
    for i in heat_utils:
        ID = i.ID
        if ID in heat_utils_dict: heat_utils_dict[ID].append(i)
        else: heat_utils_dict[ID] = [i]
    
    # Make a list of tables, keeping all results with same Type in one table
    tables = []
    for Type, heat_utils in heat_utils_dict.items():
        data = []; index = []
        for hu in heat_utils:
            data.append((source[hu].line, hu.duty, hu.flow, hu.cost))
            index.append(source[hu].ID)
        table = DataFrame(data, index=index,
                          columns=('Unit operation',
                                   'Duty (kJ/hr)',
                                   'Flow (kmol/hr)',
                                   'Cost (USD/hr)'))
        table.columns.name = Type
        tables.append(table)
    return tables
    

def power_utility_table(units): 
    """
    Return a pandas DataFrame object of power utilities.
    
    Parameters
    ----------
    units : iterable[Unit]
        
    """
    # Sort power utilities by unit type
    units = sorted(units, key=(lambda u: type(u).__name__))
    units = [u for u in units if u.power_utility]
    power_utilities = [u.power_utility for u in units]
    length = len(power_utilities)
    data = []
    for i, u, pu in zip(range(length), units, power_utilities):
        data.append((u.line, pu.rate, pu.cost))
    return DataFrame(data, index=[u.ID for u in units if u.power_utility],
                     columns=('Unit Operation', 'Rate (kW)', 'Cost (USD/hr)'))

# def lca_tables(sys):
#     all_feeds = [i for i in sys.feeds if i.characterization_factors]
#     all_products = [i for i in sys.products if i.characterization_factors]
#     net_electricity = bst.PowerUtility.sum([i.power_utility for i in sys.cost_units])
#     keys = list(net_electricity.characterization_factors)
#     keys = set(sum([list(i.characterization_factors) for i in (all_feeds + all_products)], keys))
#     tables = []
#     for key in keys:
#         feeds = [i for i in all_feeds if key in i.characterization_factors]
#         mass_flows = [i.F_mass for i in feeds]
        
#         df = DataFrame(data, index=[])
#         tables.append(df)

# %% Streams

def stream_table(streams, flow='kg/hr', percent=True, chemicals=None, **props):
    """
    Return a stream table as a pandas DataFrame object.

    Parameters
    ----------
    streams : array_like[Stream]
    flow : str
        Units for flow rate.
    props : str
        Additional stream properties and units as key-value pairs
    
    """
    
    # Prepare rows and columns
    ss = sorted(sorted([i for i in streams if i.ID], key=lambda i: i.ID), key=_stream_key)
    if not chemicals: 
        all_chemicals = tuple(set([i.chemicals for i in ss]))
        sizes = [(i, chemical.size) for i, chemical in enumerate(all_chemicals)]
        index, size = max(sizes, key=lambda x: x[1])
        chemicals = all_chemicals[index]
    n = len(ss)
    m = chemicals.size
    p = len(props)
    array = np.empty((m+p+5, n), dtype=object)
    IDs = n*[None]
    sources = array[0, :]
    sinks = array[1, :]
    phases = array[2, :]
    prop_molar_data = array[3:3+p+1,:]
    flows = array[p+3, :]
    array[p+4, :] = ''
    fracs = array[p+5:m+p+5, :]
    for j in range(n):
        s = ss[j]
        sources[j] = s.source.ID if s.source else '-'
        sinks[j] = s.sink.ID if s.sink else '-'
        IDs[j] = s.ID
        phase = ''
        for i in s.phase:
            if i == 'l':
                phase += 'liquid|'
            elif i == 'L':
                phase += 'LIQUID|'
            elif i == 'g':
                phase += 'gas|'
            elif i == 's':
                phase += 'solid|'
            elif i == 'S':
                phase += 'SOLID|'
        phase = phase.rstrip('|')
        phases[j] = phase
        flow_j = s.get_flow(units=flow)
        flows[j] = net_j = flow_j.sum()
        if percent: net_j /= 100.
        fracs_j = flow_j/net_j if net_j > 1e-24 else 0
        if s.chemicals is chemicals:
            fracs[:, j] = fracs_j
        else:
            fracs[:, j] = 0.
            fracs[chemicals.get_index(s.chemicals.IDs), j] = fracs_j
        i = 0
        for attr, units in props.items():
            prop_molar_data[i, j] = s.get_property(attr, units)
            i += 1
    index = (
        'Source', 
        'Sink',
        'Phase', 
        *[f'{attr} ({units})' for attr, units in props.items()], 
        f'flow ({flow})',
        ('Composition [%]:' if percent else 'Composition:'),
        *chemicals.IDs
    )
    return DataFrame(array, columns=IDs, index=index)





# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .._unit import Unit
from .mixing import *
from .splitting import *
from .phase_equilibrium import *
from .molecular_sieve import *
from .vacuum_system import *
from ._pump import *
from .heat_exchange import *
from .tank import *
from .distillation import *
from ._duplicator import *
from ._junction import *
from ._scaler import *
from ._balance import *
from ._diagram_only_units import *
from ._flash import *
from ._multi_effect_evaporator import *
from .solids_separation import *
from ._batch_bioreactor import *
from ._batch_crystallizer import *
from ._fermentation import *
from ._enzyme_treatment import *
from ._clarifier import *
from ._screw_feeder import *
from ._magnetic_separator import *
from ._conveying_belt import *
from ._vent_scrubber import *
from ._vibrating_screen import *
from ._carbon_capture import *
from .stirred_tank_reactor import *
from .compressor import *
from .turbine import *
from .valve import *
from .drying import *
from .size_reduction import *
from .size_enlargement import *
from .liquid_liquid_extraction import *
from .adsorption import *
from .auxiliary import *
from .agitator import *

from . import (
    _flash, 
    _pump, 
    _multi_effect_evaporator, 
    _magnetic_separator,
    _conveying_belt, 
    _vent_scrubber,
    _vibrating_screen,
    _junction,
    _scaler,
    _fermentation, 
    _enzyme_treatment, 
    _clarifier, 
    _balance,  
    _screw_feeder,
    stirred_tank_reactor,
    molecular_sieve,
    vacuum_system,
    adsorption,
    size_reduction, 
    size_enlargement,
    drying,
    distillation, 
    tank,
    liquid_liquid_extraction, 
    mixing, 
    splitting, 
    phase_equilibrium,
    heat_exchange, 
    solids_separation,
    decorators, 
    design_tools, 
    _duplicator,
    _diagram_only_units, 
    _batch_bioreactor,
    _batch_crystallizer,
    _carbon_capture,
    compressor,
    turbine,
    valve,
    auxiliary,
    agitator,
)


__all__ = ('Unit',
           *molecular_sieve.__all__,
           *liquid_liquid_extraction.__all__,
           *_diagram_only_units.__all__,
           *_flash.__all__,
           *mixing.__all__,
           *splitting.__all__,
           *phase_equilibrium.__all__,
           *_pump.__all__,
           *heat_exchange.__all__,
           *_multi_effect_evaporator.__all__,
           *distillation.__all__,
           *tank.__all__,
           *stirred_tank_reactor.__all__,
           *_conveying_belt.__all__,
           *_vent_scrubber.__all__,
           *_vibrating_screen.__all__,
           *_junction.__all__,
           *_scaler.__all__,
           *solids_separation.__all__,
           *_fermentation.__all__, 
           *_enzyme_treatment.__all__,
           *_clarifier.__all__,
           *size_reduction.__all__,
           *size_enlargement.__all__,
           *_balance.__all__, 
           *_screw_feeder.__all__,
           *_magnetic_separator.__all__,
           *_duplicator.__all__,
           *_batch_bioreactor.__all__,
           *_batch_crystallizer.__all__,
           *_carbon_capture.__all__,
           *drying.__all__,
           *adsorption.__all__,
           *compressor.__all__,
           *turbine.__all__,
           *valve.__all__,
           *vacuum_system.__all__,
           *auxiliary.__all__,
           *agitator.__all__,
           'adsorption',
           'drying',
           'tank',
           'mixing',
           'splitting',
           'phase_equilibrium',
           'distillation',
           'decorators',
           'design_tools',
           'heat_exchange',
           'solids_separation',
           'liquid_liquid_extraction',
           'size_reduction',
           'size_enlargement',
)



